import catboost as cb
import xgboost as xgb
from sklearn.ensemble import ExtraTreesClassifier, GradientBoostingClassifier, RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB, MultinomialNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.svm import SVC


from alpbench.benchmark.BenchmarkConnector import BenchmarkConnector

from alpbench.pipeline.QueryStrategy import (
    BALDQueryStrategy,
    ClusterMarginQueryStrategy,
    CoreSetQueryStrategy,
    DiscriminativeQueryStrategy,
    EntropyQueryStrategy,
    EpistemicUncertaintyQueryStrategy,
    FalcunQueryStrategy,
    KMeansQueryStrategy,
    LeastConfidentQueryStrategy,
    MarginQueryStrategy,
    MaxEntropyQueryStrategy,
    MinMarginQueryStrategy,
    MonteCarloEERLogLoss,
    MonteCarloEERMisclassification,
    PowerBALDQueryStrategy,
    PowerMarginQueryStrategy,
    QBCVarianceRatioQueryStrategy,
    QueryByCommitteeEntropyQueryStrategy,
    QueryByCommitteeKLQueryStrategy,
    RandomMarginQueryStrategy,
    RandomQueryStrategy,
    TypicalClusterQueryStrategy,
    WeightedClusterQueryStrategy
)

def ensure_default_setup(dbbc: BenchmarkConnector):
    """
    Ensures that the default settings, sampling strategies and learners are loaded into the database.
    This allows to later on load the default settings and strategies.
    """
    # init settings
    settings = [
        {
            "name": "small",
            "labeled_train_size": 30,
            "train_type": "absolute",
            "test_size": 0.33,
            "number_of_queries": 10,
            "number_of_iterations": 20,
            "factor": -1,
        },
        {
            "name": "medium",
            "labeled_train_size": 100,
            "train_type": "absolute",
            "test_size": 0.33,
            "number_of_queries": 50,
            "number_of_iterations": 20,
            "factor": -1,
        },
        {
            "name": "large",
            "labeled_train_size": 300,
            "train_type": "absolute",
            "test_size": 0.33,
            "number_of_queries": 200,
            "number_of_iterations": 20,
            "factor": -1,
        },
        {
            "name": "small_dynamic",
            "labeled_train_size": 10,
            "train_type": "absolute",
            "test_size": 0.33,
            "number_of_queries": 10,
            "number_of_iterations": 20,
            "factor": 5,
        },
        {
            "name": "large_dynamic",
            "labeled_train_size": 10,
            "train_type": "absolute",
            "test_size": 0.33,
            "number_of_queries": 10,
            "number_of_iterations": 20,
            "factor": 20,
        },
    ]
    for setting in settings:
        dbbc.load_or_create_setting(**setting)

    default_query_strategies = {
        "random": RandomQueryStrategy(42),
        "random_margin": RandomMarginQueryStrategy(42),
        "cluster_margin": ClusterMarginQueryStrategy(42),
        "core_set": CoreSetQueryStrategy(42),
        "epistemic": EpistemicUncertaintyQueryStrategy(42),
        "entropy": EntropyQueryStrategy(42),
        "falcun": FalcunQueryStrategy(42),
        "margin": MarginQueryStrategy(42),
        "max_entropy": MaxEntropyQueryStrategy(42, 10),
        "least_confident": LeastConfidentQueryStrategy(42),
        "mc_logloss": MonteCarloEERLogLoss(42),
        "mc_misclass": MonteCarloEERMisclassification(42),
        "kmeans": KMeansQueryStrategy(42),
        "discrim": DiscriminativeQueryStrategy(42),
        "qbc_entropy": QueryByCommitteeEntropyQueryStrategy(42, 10),
        "qbc_kl": QueryByCommitteeKLQueryStrategy(42, 10),
        "bald": BALDQueryStrategy(42, 10),
        "power_margin": PowerMarginQueryStrategy(42),
        "min_margin": MinMarginQueryStrategy(42, 10),
        "typ_cluster": TypicalClusterQueryStrategy(42),
        "weighted_cluster": WeightedClusterQueryStrategy(42),
        "qbc_variance_ratio": QBCVarianceRatioQueryStrategy(42, 10),
        "power_bald": PowerBALDQueryStrategy(42, 10),
    }

    for k, obj in default_query_strategies.items():
        dbbc.load_or_create_query_strategy(k, obj)

    default_learners = {
        "svm_lin": SVC(kernel="linear", probability=True),
        "svm_rbf": SVC(kernel="rbf", probability=True),
        "rf_entropy": RandomForestClassifier(n_estimators=100, max_depth=10, criterion="entropy"),
        "rf_gini": RandomForestClassifier(n_estimators=100, max_depth=10, criterion="gini"),
        "rf_entropy_large": RandomForestClassifier(n_estimators=250, max_depth=10, criterion="entropy"),
        "rf_gini_large": RandomForestClassifier(n_estimators=250, max_depth=10, criterion="gini"),
        "knn_3": KNeighborsClassifier(n_neighbors=3),
        "knn_10": KNeighborsClassifier(n_neighbors=10),
        "log_reg": LogisticRegression(),
        "multinomial_bayes": MultinomialNB(),
        "etc_entropy": ExtraTreesClassifier(n_estimators=100, max_depth=10, criterion="entropy"),
        "etc_gini": ExtraTreesClassifier(n_estimators=100, max_depth=10, criterion="gini"),
        "etc_entropy_large": ExtraTreesClassifier(n_estimators=250, max_depth=10, criterion="entropy"),
        "etc_gini_large": ExtraTreesClassifier(n_estimators=250, max_depth=10, criterion="gini"),
        "naive_bayes": GaussianNB(),
        "mlp": MLPClassifier(),
        "GBT_logloss": GradientBoostingClassifier(n_estimators=100),
        "GBT_exp": GradientBoostingClassifier(n_estimators=100, loss="exponential"),
        "GBT_logloss_large": GradientBoostingClassifier(n_estimators=250),
        "xgb": xgb.XGBClassifier(tree_method="hist", max_depth=6, n_estimators=100),
        "catboost": cb.CatBoostClassifier(rsm=0.1, iterations=100, depth=10, verbose=0, boosting_type='Plain',
                                          bootstrap_type='Bernoulli', subsample=.75),
    }

    for k, obj in default_learners.items():
        dbbc.load_or_create_learner(k, obj)



import json
from abc import ABC, abstractmethod
from pathlib import Path

import mysql.connector

from alpbench.benchmark.ActiveLearningScenario import ActiveLearningScenario
from alpbench.benchmark.ActiveLearningSetting import ActiveLearningSetting
from alpbench.util.common import format_insert_query, format_select_query, fullname, instantiate_class_by_fqn


class BenchmarkConnector(ABC):
    """Benchmark Connector

    This abstract class defines the interface for a benchmark connector. A benchmark connector is responsible for
    storing and loading all data related to the active learning benchmark. This includes the active learning setting and
    scenario as well as the learner and query strategy with all used parameters. The data is stored in a database or
    file and can be accessed by the respective methods of this class.

    """

    def __init__(self):
        pass

    @abstractmethod
    def load_scenario(self, scenario_id: int):
        """
        Abstract method that loads the scenario with the specified ID from the database.
        """
        pass

    @abstractmethod
    def load_or_create_scenario(self, openml_id, test_split_seed, train_split_seed, seed, setting_id):
        """
        Abstract method that loads the scenario with the specified parameters from the database. If the scenario does
        not exist yet, it is created and then returned to the invoker.
        """
        pass

    @abstractmethod
    def load_setting(self, setting_id: int):
        """
        Abstract method that loads the setting with the specified ID from the database.
        """
        pass

    @abstractmethod
    def load_setting_by_name(self, setting_name):
        """
        Abstract method that loads the setting with the specified name from the database.
        """
        pass

    @abstractmethod
    def load_or_create_setting(
        self, name, labeled_train_size, train_type, test_size, number_of_iterations, number_of_queries, factor
    ):
        """
        Abstract method that loads the setting with the specified parameters from the database. If the setting does
        not exist yet, it is created and then returned to the invoker.
        """
        pass

    @abstractmethod
    def load_learner_by_name(self, learner_name):
        """
        Abstract method that loads the learner with the specified name from the database.
        """
        pass

    @abstractmethod
    def load_learner(self, learner_id: int):
        """
        Abstract method that loads the learner with the specified ID from the database.
        """
        pass

    @abstractmethod
    def load_or_create_learner(self, learner_name, obj):
        """
        Abstract method that loads the learner with the specified parameters from the database. If the learner does
        not exist yet, it is created and then returned to the invoker.
        """
        pass

    @abstractmethod
    def load_query_strategy_by_name(self, query_strategy_name):
        """
        Abstract method that loads the query strategy with the specified name from the database.
        """
        pass

    @abstractmethod
    def load_query_strategy(self, query_strategy_id):
        """
        Abstract method that loads the query strategy with the specified ID from the database.
        """
        pass

    @abstractmethod
    def load_or_create_query_strategy(self, query_strategy_name, obj):
        """
        Abstract method that loads the query strategy with the specified parameters from the database. If the
        query strategy does not exist yet, it is created and then returned to the invoker.
        """
        pass


def _fetch_data_of_descriptor(data, descriptor: dict):
    """
    This method fetches the data from the specified list that matches the specified descriptor. The descriptor is a
    dictionary that contains key-value pairs that must match the data in order to be fetched.

    Parameters:
        data (list): list of dictionaries that contain the data
        descriptor (dict): dictionary that contains the key-value pairs that must match the data

    Returns:
        dict: the data that matches the descriptor
    """
    fetched_entry = None
    for entry in data:
        does_match = True
        for k, v in descriptor.items():
            if entry[k] != descriptor[k]:
                does_match = False
                break
        if does_match:
            fetched_entry = entry
            break
    return fetched_entry


class DataFileBenchmarkConnector(BenchmarkConnector):
    """Data File Benchmark Connector

    This class is an implementation of the BenchmarkConnector interface that stores all data in files. This involves
    the Active Learning Setting and Scenario as well as the Learner and query Strategy with all used parameters.
    The data is stored in JSON files and can be accessed by the respective methods of this class.

    Args:
        learner_file (str): path to the file storing the learner data
        query_strategy_file (str): path to the file storing the query strategy data
        scenario_file (str): path to the file storing the scenario data
        setting_file (str): path to the file storing the setting data

    Attributes:
        learner_file (str): path to the file storing the learner data
        query_strategy_file (str): path to the file storing the query strategy data
        scenario_file (str): path to the file storing the scenario data
        setting_file (str): path to the file storing the setting data

    """

    base_folder = "db_alpbench/"
    learner_file = base_folder + "learner.json"
    query_strategy_file = base_folder + "query_strategy.json"
    scenario_file = base_folder + "scenario.json"
    setting_file = base_folder + "setting.json"

    def __init__(self, learner_file=None, query_strategy_file=None, scenario_file=None, setting_file=None):
        self.learner_file = learner_file if learner_file is not None else DataFileBenchmarkConnector.learner_file
        self.query_strategy_file = (
            query_strategy_file if learner_file is not None else DataFileBenchmarkConnector.query_strategy_file
        )
        self.scenario_file = scenario_file if scenario_file is not None else DataFileBenchmarkConnector.scenario_file
        self.setting_file = setting_file if setting_file is not None else DataFileBenchmarkConnector.setting_file

        # ensure files exist and are at least empty json arrays
        import os

        for file in [self.scenario_file, self.setting_file, self.learner_file, self.query_strategy_file]:
            p = Path(file)
            os.makedirs(p.parent, exist_ok=True)

            if not os.path.isfile(file):
                with open(file, "w") as f:
                    f.write("[]")
        with open(self.scenario_file) as f:
            self.scenarios = json.load(f)
        with open(self.setting_file) as f:
            self.settings = json.load(f)
        with open(self.learner_file) as f:
            self.learners = json.load(f)
        with open(self.query_strategy_file) as f:
            self.query_strategies = json.load(f)

    def cleanup(self):
        self.dump()

    def dump(self):
        dump_data = [
            (self.setting_file, self.settings),
            (self.scenario_file, self.scenarios),
            (self.learner_file, self.learners),
            (self.query_strategy_file, self.query_strategies),
        ]
        for dd in dump_data:
            with open(dd[0], "w") as f:
                json.dump(obj=dd[1], fp=f, indent=4)

    def load_scenario(self, scenario_id: int):
        """
        This method loads the scenario with the specified ID from the database.

        Parameters:
            scenario_id (int): ID of the scenario to load

        Returns:
            ActiveLearningScenario: the loaded scenario
        """
        stored_scenario = _fetch_data_of_descriptor(self.scenarios, {"scenario_id": scenario_id})
        if stored_scenario is None:
            raise BaseException("No scenario could be found with ID " + str(scenario_id))

        setting = self.load_setting(stored_scenario.pop("setting_id"))
        stored_scenario["setting"] = setting

        return ActiveLearningScenario(**stored_scenario)

    def load_or_create_scenario(self, openml_id, test_split_seed, train_split_seed, seed, setting_id):
        """
        This method loads the scenario with the specified parameters from the database. If the scenario does not exist
        yet, it is created and then returned to the invoker.

        Parameters:
            openml_id (int): ID of the openml dataset
            test_split_seed (int): seed for the test split
            train_split_seed (int): seed for the train split
            seed (int): seed for the scenario
            setting_id (int): ID of the setting

        Returns:
            ActiveLearningScenario: the loaded or created scenario
        """
        descriptor = {
            "openml_id": openml_id,
            "test_split_seed": test_split_seed,
            "train_split_seed": train_split_seed,
            "seed": seed,
            "setting_id": setting_id,
        }
        stored_scenario = _fetch_data_of_descriptor(self.scenarios, descriptor)

        setting = self.load_setting(descriptor.pop("setting_id"))
        descriptor["setting"] = setting

        if stored_scenario is None:
            max_id = 0
            for s in self.scenarios:
                max_id = max(s["scenario_id"], max_id)
            descriptor["scenario_id"] = max_id + 1

            als = ActiveLearningScenario(**descriptor)

            descriptor["labeled_indices"] = als.get_labeled_instances()
            descriptor["test_indices"] = als.get_test_indices()
            descriptor["setting_id"] = setting.get_setting_id()
            descriptor.pop("setting")
            self.scenarios += [dict(descriptor)]

            return als
        else:
            descriptor["scenario_id"] = stored_scenario["scenario_id"]
            return ActiveLearningScenario(**descriptor)

    def load_setting(self, setting_id):
        """
        This method loads the setting with the specified ID from the database.

        Parameters:
            setting_id (int): ID of the setting to load

        Returns:
            ActiveLearningSetting: the loaded setting
        """
        stored_setting = _fetch_data_of_descriptor(self.settings, {"setting_id": setting_id})

        if stored_setting is None:
            raise BaseException("No setting could be found with ID " + str(setting_id))

        return ActiveLearningSetting(**stored_setting)

    def load_setting_by_name(self, setting_name):
        """
        This method loads the setting with the specified name from the database.

        Parameters:
            setting_name (str): name of the setting to load

        Returns:
            ActiveLearningSetting: the loaded setting
        """
        stored_setting = _fetch_data_of_descriptor(self.settings, {"setting_name": setting_name})

        if stored_setting is None:
            raise BaseException("No setting could be found with name " + setting_name)

        return ActiveLearningSetting(**stored_setting)

    def load_or_create_setting(
        self, name, labeled_train_size, train_type, test_size, number_of_iterations, number_of_queries, factor
    ):
        """
        This method checks whether the specified setting already exists. If so, it just fetches the data from the
        database and returns an instance. If not, the specified setting is added to the database and then also returned.

        Parameters:
            name (str): name of the setting
            labeled_train_size (str): labeled training size
            train_type (str): type of training
            test_size (str): test size
            number_of_iterations (int): number of iterations
            number_of_queries (int): number of queries
            factor (int): factor

        Returns:
            ActiveLearningSetting: the loaded or created setting
        """
        setting_descriptor = {
            "setting_name": name,
            "setting_labeled_train_size": labeled_train_size,
            "setting_train_type": train_type,
            "setting_test_size": test_size,
            "number_of_iterations": number_of_iterations,
            "number_of_queries": number_of_queries,
            "factor": factor,
        }

        stored_setting = _fetch_data_of_descriptor(self.settings, setting_descriptor)

        if stored_setting is None:
            max_id = 0
            for s in self.settings:
                max_id = max(s["setting_id"], max_id)

            setting_descriptor["setting_id"] = max_id + 1
            self.settings += [setting_descriptor]
            return ActiveLearningSetting(**setting_descriptor)
        else:
            return ActiveLearningSetting(**stored_setting)

    def load_learner_by_name(self, learner_name):
        """
        This method loads the learner with the specified name from the database.

        Parameters:
            learner_name (str): name of the learner to load

        Returns:
            object: the loaded learner
        """
        stored_learner = _fetch_data_of_descriptor(self.learners, {"learner_name": learner_name})

        if stored_learner is None:
            raise BaseException("No learner could be found with name " + learner_name)

        return instantiate_class_by_fqn(
            stored_learner["learner_class"], json.loads(stored_learner["learner_parameterization"])
        )

    def load_learner(self, learner_id):
        """
        This method loads the learner with the specified ID from the database.

        Parameters:
            learner_id (int): ID of the learner to load

        Returns:
            object: the loaded learner
        """
        stored_learner = _fetch_data_of_descriptor(self.learners, {"learner_id": learner_id})

        if stored_learner is None:
            raise BaseException("No learner could be found with ID " + str(learner_id))

        return instantiate_class_by_fqn(
            stored_learner["learner_class"], json.loads(stored_learner["learner_parameterization"])
        )

    def load_or_create_learner(self, learner_name, obj):
        """
        This method checks whether the specified learner already exists in the database. If not, the specified setting
        is added to the database and then also returned to the invoker.

        Parameters:
            learner_name (str): name of the learner
            obj (object): the learner to load or create

        Returns:
            object: the loaded or created learner
        """
        data = obj.get_params()

        learner_descriptor = {
            "learner_class": fullname(obj),
            "learner_parameterization": json.dumps(data, cls=CustomEncoder),
        }

        # check whether the specified setting already exists. if so, fetch its id from the database and return an
        # instance of that setting
        stored_learner = _fetch_data_of_descriptor(self.learners, learner_descriptor)

        if stored_learner is None:
            # The specified setting does not yet exist so create it in the database and then return it to the invoker.
            learner_descriptor["learner_name"] = learner_name

            max_id = 0
            for learner in self.learners:
                max_id = max(learner["learner_id"], max_id)

            learner_descriptor["learner_id"] = max_id + 1
            self.learners += [learner_descriptor]
        else:
            learner_descriptor["learner_name"] = stored_learner["learner_name"]

        return learner_descriptor["learner_name"], obj

    def load_query_strategy_by_name(self, query_strategy_name):
        """
        This method loads the query strategy with the specified name from the database.

        Parameters:
            query_strategy_name (str): name of the query strategy to load

        Returns:
            object: the loaded query strategy
        """
        stored_query_strategy = _fetch_data_of_descriptor(
            self.query_strategies, {"query_strategy_name": query_strategy_name}
        )

        if stored_query_strategy is None:
            raise BaseException("No query strategy could be found with ID " + query_strategy_name)

        return instantiate_class_by_fqn(
            stored_query_strategy["query_strategy_class"],
            json.loads(stored_query_strategy["query_strategy_parameterization"]),
        )

    def load_query_strategy(self, query_strategy_id):
        """
        This method loads the query strategy with the specified ID from the database.

        Parameters:
            query_strategy_id (int): ID of the query strategy to load

        Returns:
            object: the loaded query strategy
        """
        stored_query_strategy = _fetch_data_of_descriptor(
            self.query_strategies, {"query_strategy_id": query_strategy_id}
        )

        if stored_query_strategy is None:
            raise BaseException("No query strategy could be found with ID " + str(query_strategy_id))

        return instantiate_class_by_fqn(
            stored_query_strategy["query_strategy_class"],
            json.loads(stored_query_strategy["query_strategy_parameterization"]),
        )

    def load_or_create_query_strategy(self, query_strategy_name, obj):
        """
        This method checks whether the specified query strategy already exists in the database. If not, the specified
        query strategy including its parameterization is added to the database and then also returned to the invoker.

        Parameters:
            query_strategy_name (str): name of the query strategy
            obj (object): the query strategy to load or create

        Returns:
            object: the loaded or created query strategy
        """
        query_strategy_descriptor = {
            "query_strategy_class": fullname(obj),
            "query_strategy_parameterization": json.dumps(obj.get_params()),
        }
        stored_query_strategy = _fetch_data_of_descriptor(self.query_strategies, query_strategy_descriptor)

        # check whether the specified setting already exists. if so, fetch its id from the database and return an
        # instance of that setting

        if stored_query_strategy is None:
            # The specified setting does not yet exist so create it in the database and then return it to the invoker.
            max_id = 0
            for query_strategy in self.query_strategies:
                max_id = max(query_strategy["query_strategy_id"], max_id)

            query_strategy_descriptor["query_strategy_id"] = max_id + 1
            query_strategy_descriptor["query_strategy_name"] = query_strategy_name
            self.query_strategies += [query_strategy_descriptor]
        else:
            query_strategy_descriptor["query_strategy_name"] = stored_query_strategy["query_strategy_name"]

        return query_strategy_descriptor["query_strategy_name"], obj


class MySQLBenchmarkConnector(BenchmarkConnector):
    """MySQL Benchmark Connector

    This class is an implementation of the BenchmarkConnector interface that stores all data in a MySQL database. This
    involves the Active Learning Setting and Scenario as well as the Learner and query Strategy with all used
    parameters. The data is stored in MySQL tables and can be accessed by the respective methods of this class.

    Args:
        host (str): host of the MySQL server
        user (str): user of the MySQL server
        password (str): password of the MySQL server
        database (str): database to use
        use_ssl (bool): whether to use SSL

    Attributes:
        host (str): host of the MySQL server
        user (str): user of the MySQL server
        password (str): password of the MySQL server
        database (str): database to use
        use_ssl (bool): whether to use SSL
        con (mysql.connector.connection.MySQLConnection): connection to the MySQL database
        scenario_table (str): name of the table storing the scenarios
        setting_table (str): name of the table storing the settings
        learner_table (str): name of the table storing the learners
        query_strategy_table (str): name of the table storing the query strategies
    """

    scenario_table = "salt_scenario"
    setting_table = "salt_setting"
    learner_table = "salt_learner"
    query_strategy_table = "salt_query_strategy"

    def __init__(self, host, user, password, database, use_ssl):
        super().__init__()
        self.host = host
        self.user = user
        self.password = password
        self.database = database
        self.use_ssl = use_ssl

        self.con = mysql.connector.connect(
            host=host, user=user, password=password, database=database, ssl_disabled=not use_ssl
        )

        setting_table_query = (
            f"CREATE TABLE IF NOT EXISTS {MySQLBenchmarkConnector.setting_table} ("
            f"setting_id INT(10) NOT NULL AUTO_INCREMENT PRIMARY KEY, "
            f"setting_name VARCHAR(250) UNIQUE, setting_labeled_train_size VARCHAR(50), "
            f"setting_train_type VARCHAR(250), setting_test_size VARCHAR(50), "
            f"number_of_iterations INT(10), number_of_queries INT(10), factor INT(10))"
        )
        scenario_table_query = (
            f"CREATE TABLE IF NOT EXISTS {MySQLBenchmarkConnector.scenario_table} ("
            f"scenario_id INT(10) NOT NULL AUTO_INCREMENT PRIMARY KEY, openml_id INT(10), "
            f"test_split_seed INT(10), train_split_seed INT(10), seed INT(10), "
            f"labeled_indices LONGTEXT, test_indices LONGTEXT, setting_id INT(10))"
        )
        learner_table_query = (
            f"CREATE TABLE IF NOT EXISTS {MySQLBenchmarkConnector.learner_table} ("
            f"learner_id INT(10) NOT NULL AUTO_INCREMENT PRIMARY KEY, "
            f"learner_name VARCHAR(100) UNIQUE, learner_class VARCHAR(250), "
            f"learner_parameterization TEXT)"
        )
        query_strategy_table_q = (
            f"CREATE TABLE IF NOT EXISTS {MySQLBenchmarkConnector.query_strategy_table}"
            f" (query_strategy_id INT(10) NOT NULL AUTO_INCREMENT PRIMARY KEY, "
            f"query_strategy_name VARCHAR(100) UNIQUE,"
            f"query_strategy_class VARCHAR(250), "
            f"query_strategy_parameterization TEXT)"
        )

        cursor = self.con.cursor()
        for q in [setting_table_query, scenario_table_query, learner_table_query, query_strategy_table_q]:
            cursor.execute(q)

    def close(self):
        self.con.close()

    def load_scenario(self, scenario_id):
        """
        This method loads the scenario with the specified ID from the database.

        Parameters:
            scenario_id (int): ID of the scenario to load

        Returns:
            ActiveLearningScenario: the loaded scenario
        """
        query = format_select_query(MySQLBenchmarkConnector.scenario_table, {"scenario_id": scenario_id})
        cursor = self.con.cursor(buffered=True, dictionary=True)
        cursor.execute(query)
        res = cursor.fetchall()
        if len(res) > 0:
            scenario_data = res[0]
            setting = self.load_setting(res[0]["setting_id"])
            scenario_data.pop("setting_id")
            scenario_data["setting"] = setting
            scenario_data["labeled_indices"] = json.loads(scenario_data["labeled_indices"])
            scenario_data["test_indices"] = json.loads(scenario_data["test_indices"])
            return ActiveLearningScenario(**scenario_data)
        else:
            raise Exception("Scenario with ID " + str(scenario_id) + " unknown")

    def load_or_create_scenario(self, openml_id, test_split_seed, train_split_seed, seed, setting_id):
        """
        This method loads the scenario with the specified parameters from the database. If the scenario does not exist
        yet, it is created and then returned to the invoker.

        Parameters:
            openml_id (int): ID of the openml dataset
            test_split_seed (int): seed for the test split
            train_split_seed (int): seed for the train split
            seed (int): seed for the scenario
            setting_id (int): ID of the setting

        Returns:
            ActiveLearningScenario: the loaded or created scenario
        """
        setting = self.load_setting(setting_id)
        scenario_data = {
            "openml_id": openml_id,
            "test_split_seed": test_split_seed,
            "train_split_seed": train_split_seed,
            "seed": seed,
            "setting_id": setting_id,
        }
        check_query = format_select_query(MySQLBenchmarkConnector.scenario_table, scenario_data)
        cursor = self.con.cursor(buffered=True, dictionary=True)
        cursor.execute(check_query)
        res_check = cursor.fetchall()
        cursor.close()

        scenario_data["setting"] = setting
        scenario_data.pop("setting_id")

        if len(res_check) > 0:
            scenario_data["scenario_id"] = int(res_check[0]["scenario_id"])
            return self.load_scenario(scenario_data["scenario_id"])

        scenario_data["scenario_id"] = -1
        new_scenario = ActiveLearningScenario(**scenario_data)
        scenario_data.pop("setting")
        scenario_data.pop("scenario_id")
        scenario_data["setting_id"] = setting_id
        scenario_data["labeled_indices"] = str(new_scenario.get_labeled_instances())
        scenario_data["test_indices"] = str(new_scenario.get_test_indices())

        insert_query = format_insert_query(MySQLBenchmarkConnector.scenario_table, scenario_data)
        cursor = self.con.cursor()
        cursor.execute(insert_query)
        scenario_id = cursor.lastrowid
        self.con.commit()

        new_scenario.scenario_id = scenario_id
        return new_scenario

    def load_setting(self, setting_id):
        """
        This method loads the setting with the specified ID from the database.

        Parameters:
            setting_id (int): ID of the setting to load

        Returns:
            ActiveLearningSetting: the loaded setting
        """
        query = format_select_query(MySQLBenchmarkConnector.setting_table, {"setting_id": setting_id})
        cursor = self.con.cursor(buffered=True, dictionary=True)
        cursor.execute(query)
        res = cursor.fetchall()
        if len(res) > 0:
            return ActiveLearningSetting(**res[0])
        else:
            raise Exception("Setting with ID " + str(setting_id) + " could not be found.")

    def load_setting_by_name(self, setting_name):
        """
        This method loads the setting with the specified name from the database.

        Parameters:
            setting_name (str): name of the setting to load

        Returns:
            ActiveLearningSetting: the loaded setting
        """
        query = format_select_query(MySQLBenchmarkConnector.setting_table, {"setting_name": setting_name})
        cursor = self.con.cursor(buffered=True, dictionary=True)
        cursor.execute(query)
        res = cursor.fetchall()
        if len(res) > 0:
            return ActiveLearningSetting(**res[0])
        else:
            raise Exception("Setting with name " + str(setting_name) + " could not be found.")

    def load_or_create_setting(
        self, name, labeled_train_size, train_type, test_size, number_of_iterations, number_of_queries, factor
    ):
        """
        This method checks whether the specified setting already exists. If so, it just fetches the data from the
        database and returns an instance. If not, the specified setting is added to the database and then also returned
        to the invoker.

        Parameters:
            name (str): name of the setting
            labeled_train_size (str): labeled training size
            train_type (str): type of training
            test_size (str): test size
            number_of_iterations (int): number of iterations
            number_of_queries (int): number of queries
            factor (int): factor

        Returns:
            ActiveLearningSetting: the loaded or created setting
        """
        setting_descriptor = {
            "setting_name": name,
            "setting_labeled_train_size": labeled_train_size,
            "setting_train_type": train_type,
            "setting_test_size": test_size,
            "number_of_iterations": number_of_iterations,
            "number_of_queries": number_of_queries,
            "factor": factor,
        }

        # check whether the specified setting already exists. if so, fetch its id from the database and return an
        # instance of that setting
        query_check = format_select_query(MySQLBenchmarkConnector.setting_table, setting_descriptor)
        cursor = self.con.cursor(buffered=True, dictionary=True)
        cursor.execute(query_check)
        res_check = cursor.fetchall()
        cursor.close()

        if len(res_check) > 0:
            setting_descriptor["setting_id"] = res_check[0]["setting_id"]
            return ActiveLearningSetting(**setting_descriptor)

        # The specified setting does not yet exist so create it in the database and then return it to the invoker.
        query = format_insert_query(MySQLBenchmarkConnector.setting_table, setting_descriptor)
        cursor = self.con.cursor()
        cursor.execute(query)
        inserted_id = cursor.lastrowid
        self.con.commit()

        return ActiveLearningSetting(
            setting_id=inserted_id,
            setting_name=name,
            setting_test_size=test_size,
            setting_train_type=train_type,
            setting_labeled_train_size=labeled_train_size,
            number_of_iterations=number_of_iterations,
            number_of_queries=number_of_queries,
            factor=factor,
        )

    def load_learner_by_name(self, learner_name):
        """
        This method loads the learner with the specified name from the database.

        Parameters:
            learner_name (str): name of the learner to load

        Returns:
            object: the loaded learner
        """
        query = format_select_query(MySQLBenchmarkConnector.learner_table, {"learner_name": learner_name})
        cursor = self.con.cursor(buffered=True, dictionary=True)
        cursor.execute(query)
        res = cursor.fetchall()
        if len(res) > 0:
            learner_data = res[0]
            return instantiate_class_by_fqn(
                learner_data["learner_class"], json.loads(learner_data["learner_parameterization"])
            )
        else:
            raise Exception("Learner with name " + str(learner_name) + " unknown")

    def load_learner(self, learner_id):
        """
        This method loads the learner with the specified ID from the database.

        Parameters:
            learner_id (int): ID of the learner to load

        Returns:
            object: the loaded learner
        """
        query = format_select_query(MySQLBenchmarkConnector.learner_table, {"learner_id": learner_id})
        cursor = self.con.cursor(buffered=True, dictionary=True)
        cursor.execute(query)
        res = cursor.fetchall()
        if len(res) > 0:
            learner_data = res[0]
            return instantiate_class_by_fqn(
                learner_data["learner_class"], json.loads(learner_data["learner_parameterization"])
            )
        else:
            raise Exception("Learner with ID " + str(learner_id) + " unknown")

    def load_or_create_learner(self, learner_name, obj):
        """
        This method checks whether the specified learner already exists in the database. If not, the specified setting
        is added to the database and then also returned to the invoker.

        Parameters:
            learner_name (str): name of the learner
            obj (object): the learner to load or create

        Returns:
            object: the loaded or created learner
        """
        learner_descriptor = {
            "learner_class": fullname(obj),
            "learner_parameterization": json.dumps(obj.get_params()),
        }

        # check whether the specified setting already exists. if so, fetch its id from the database and return an
        # instance of that setting
        query_check = format_select_query(MySQLBenchmarkConnector.learner_table, learner_descriptor)
        cursor = self.con.cursor(buffered=True, dictionary=True)
        cursor.execute(query_check)
        res_check = cursor.fetchall()
        cursor.close()

        if len(res_check) < 1:
            # The specified setting does not yet exist so create it in the database and then return it to the invoker.
            learner_descriptor["learner_name"] = learner_name
            query = format_insert_query(MySQLBenchmarkConnector.learner_table, learner_descriptor)
            cursor = self.con.cursor()
            cursor.execute(query)
            self.con.commit()
        else:
            learner_descriptor["learner_name"] = res_check[0]["learner_name"]

        return learner_descriptor["learner_name"], obj

    def load_query_strategy_by_name(self, query_strategy_name):
        """
        This method loads the query strategy with the specified name from the database.

        Parameters:
            query_strategy_name (str): name of the query strategy to load

        Returns:
            object: the loaded query strategy
        """
        query = format_select_query(
            MySQLBenchmarkConnector.query_strategy_table, {"query_strategy_name": query_strategy_name}
        )
        cursor = self.con.cursor(buffered=True, dictionary=True)
        cursor.execute(query)
        res = cursor.fetchall()
        if len(res) > 0:
            query_strategy_data = res[0]
            return instantiate_class_by_fqn(
                query_strategy_data["query_strategy_class"],
                json.loads(query_strategy_data["query_strategy_parameterization"]),
            )
        else:
            raise Exception("query strategy with name " + str(query_strategy_name) + " unknown")

    def load_query_strategy(self, query_strategy_id):
        """
        This method loads the query strategy with the specified ID from the database.

        Parameters:
            query_strategy_id (int): ID of the query strategy to load

        Returns:
            object: the loaded query strategy
        """
        query = format_select_query(
            MySQLBenchmarkConnector.query_strategy_table, {"query_strategy_id": query_strategy_id}
        )
        cursor = self.con.cursor(buffered=True, dictionary=True)
        cursor.execute(query)
        res = cursor.fetchall()
        if len(res) > 0:
            query_strategy_data = res[0]
            return instantiate_class_by_fqn(
                query_strategy_data["query_strategy_class"],
                json.loads(query_strategy_data["query_strategy_parameterization"]),
            )
        else:
            raise Exception("query strategy with ID " + str(query_strategy_id) + " unknown")

    def load_or_create_query_strategy(self, query_strategy_name, obj):
        """
        This method checks whether the specified query strategy already exists in the database. If not, the specified
        query strategy including its parameterization is added to the database and then also returned to the invoker.

        Parameters:
            query_strategy_name (str): name of the query strategy
            obj (object): the query strategy to load or create

        Returns:
            object: the loaded or created query strategy
        """
        query_strategy_descriptor = {
            "query_strategy_class": fullname(obj),
            "query_strategy_parameterization": json.dumps(obj.get_params()),
        }

        # check whether the specified setting already exists. if so, fetch its id from the database and return an
        # instance of that setting
        query_check = format_select_query(MySQLBenchmarkConnector.query_strategy_table, query_strategy_descriptor)
        cursor = self.con.cursor(buffered=True, dictionary=True)
        cursor.execute(query_check)
        res_check = cursor.fetchall()
        cursor.close()

        if len(res_check) < 1:
            # The specified setting does not yet exist so create it in the database and then return it to the invoker.
            query_strategy_descriptor["query_strategy_name"] = query_strategy_name
            query = format_insert_query(MySQLBenchmarkConnector.query_strategy_table, query_strategy_descriptor)
            cursor = self.con.cursor()
            cursor.execute(query)
            self.con.commit()
        else:
            query_strategy_descriptor["query_strategy_name"] = res_check[0]["query_strategy_name"]

        return query_strategy_descriptor["query_strategy_name"], obj


class CustomEncoder(json.JSONEncoder):
    """CustomEncoder

    This class is a custom JSON encoder that is used to encode the parameters of a learner or query strategy.
    """

    def default(self, obj):
        """
        This method is called by the JSON encoder to encode the specified object.

        Parameters:
            obj (object): object to encode

        Returns:
            str: the encoded object
        """
        if isinstance(obj, type):
            return str(obj)  # Convert class references to their string representation
        return json.JSONEncoder.default(self, obj)



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.colors import Normalize
import pandas as pd
from py_experimenter.experimenter import PyExperimenter
import json
import os
import sqlite3
from scipy.stats import ttest_ind_from_stats
from scipy.stats import ttest_ind as tt


class StudyData():

    def __init__(self, dir_to_save):
        self.dir_to_save = dir_to_save
        if not os.path.exists(dir_to_save + "/"):
            os.makedirs(dir_to_save + "/")
        self.base = None

    def generate_summary_df(self):
        results = pd.read_csv(self.dir_to_save+"/results.csv")
        if self.base == "file":
            results = results.rename(columns={'ID': 'experiment_id'})
        accuracies = pd.read_csv(self.dir_to_save+"/accuracies.csv")
        labeling = pd.read_csv(self.dir_to_save+"/labeling.csv")

        experiment_ids = results['experiment_id'].values

        data = accuracies['model_dict'].values
        max_iter = 20

        exp_ids = []
        for i in experiment_ids[:]:
            if i in labeling['experiment_id'].values and i in accuracies['experiment_id'].values:
                exp_ids.append(i)

        dict_to_fill = pd.DataFrame(columns=['experiment_id', 'iteration', 'test_accuracy', 'test_f1',
                                                'test_precision', 'test_recall', 'test_auc', 'test_log_loss',
                                                'len_X_l'])

        for enum,i in enumerate(exp_ids[:]):
            if enum%100 == 0:
                print("Experiment", enum, "of", len(exp_ids))
            for j in range(int(max_iter) + 1):
                lab = labeling[labeling['experiment_id'] == i]
                label = lab['data_dict'].values
                actual_max_iter = max(map(int, json.loads(data[0]).keys()))
                if j > actual_max_iter:
                    continue
                label_dict = json.loads(label[0])[str(j)]
                acc = accuracies[accuracies['experiment_id'] == i]
                data = acc['model_dict'].values
                dict = json.loads(data[0])[str(j)]
                dict_to_fill.loc[len(dict_to_fill)] = [int(i), int(dict['iteration']), dict['test_accuracy'],
                                                        dict['test_f1'],
                                                        dict['test_precision'], dict['test_recall'],
                                                        dict['test_auc'], dict['test_log_loss'],
                                                        int(label_dict['len_X_l'])]

        merged_df = pd.merge(results, dict_to_fill, on='experiment_id', how='inner')
        merged_df = merged_df[['setting_name', 'openml_id', 'learner_name', 'query_strategy_name',
                                'test_split_seed', 'train_split_seed', 'seed', 'iteration', 'test_accuracy',
                                'test_f1',
                                'test_precision', 'test_recall', 'test_auc', 'test_log_loss', 'len_X_l']]
        end_df = merged_df
        end_df.to_csv(self.dir_to_save + "/summarized_df.csv")
        return end_df


    def generate_aubc_df(self, df):
        aubc_df = pd.DataFrame(columns=['setting_name', 'openml_id', 'learner_name', 'query_strategy_name',
                                'test_split_seed', 'train_split_seed', 'seed', "aubc"])
        for openmlid in df['openml_id'].unique():
            for learner_name in df['learner_name'].unique():
                for query_strategy_name in df['query_strategy_name'].unique():
                    for test_split_seed in df['test_split_seed'].unique():
                        for train_split_seed in df['train_split_seed'].unique():
                            for seed in df['seed'].unique():
                                sub_df = df[(df['openml_id'] == openmlid) & (df['learner_name'] == learner_name) &
                                            (df['query_strategy_name'] == query_strategy_name) &
                                            (df['test_split_seed'] == test_split_seed) &
                                            (df['train_split_seed'] == train_split_seed) &
                                            (df['seed'] == seed)]
                                if len(sub_df) == 0:
                                    continue
                                aubc = np.trapz(sub_df['test_accuracy'].values, sub_df['len_X_l'].values)
                                aubc_df.loc[len(aubc_df)] = [sub_df['setting_name'].values[0], openmlid, learner_name,
                                                            query_strategy_name, test_split_seed, train_split_seed, seed, aubc]

        aubc_df.to_csv(self.dir_to_save + "/aubc_df.csv")

class StudyDataFromFile(StudyData):

    def __init__(self, dir_to_save, source):
            super().__init__(dir_to_save=dir_to_save)
            self.source = source
            self.base = "file"


    def get_dataframes(self):
        # Connect to the database
        conn = sqlite3.connect(self.source)
        # Create a cursor object to interact with the database
        cursor = conn.cursor()

        # Get the list of tables in the database
        cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
        tables = cursor.fetchall()

        table_name_dict = {"results": "results", "results__accuracy_log": "accuracies",
                           "results__labeling_log": "labeling"}

        # Display the contents of each table
        for table_name in tables:
            table_name = table_name[0]
            query = f"SELECT * FROM {table_name}"
            df = pd.read_sql_query(query, conn)
            if table_name in table_name_dict.keys():
                df.to_csv(self.dir_to_save + "/" + table_name_dict[table_name] + ".csv")



class StudyDataFromDatabase(StudyData):
    def __init__(self, dir_to_save, db_config, scenario_config):
        super().__init__(dir_to_save=dir_to_save)
        self.db_config, self.scenario_config = db_config, scenario_config
        self.base = "database"

    def get_dataframes(self):
        experimenter_scenarios = PyExperimenter(experiment_configuration_file_path=self.scenario_config,
                                                database_credential_file_path=self.db_config)
        results = experimenter_scenarios.get_table()
        results = results.rename(columns={'ID': 'experiment_id'})
        accuracies = experimenter_scenarios.get_logtable('accuracy_log')
        labeling = experimenter_scenarios.get_logtable('labeling_log')

        results.to_csv(self.dir_to_save+"/results.csv")
        accuracies.to_csv(self.dir_to_save+"/accuracies.csv")
        labeling.to_csv(self.dir_to_save+"/labeling.csv")


from alpbench.util.pytorch_tabnet.callbacks import Callback
import torch
import torch.nn.functional as F
import torch.nn as nn
import time

from alpbench.util.transformer import TransformerModel
from alpbench.util.transformer_prediction_interface import TabPFNClassifier

class TabPFNEmbedder(nn.Module):
    """TabPFNEmbedder

    This class is used to create a TabPFNEmbedder model. The model is used to encode the input data into a
    lower-dimensional space.

    Args:
        X_train: numpy.ndarray
        y_train: numpy.ndarray

    Attributes:
        clf: object
        num_samples: int
        encoder: object
        fc1: object
        drop: object
        fc2: object
    """
    def __init__(self, X_train, y_train):
        super().__init__()
        self.clf = None
        self.num_samples = None
        self.instantiate_tabpfn(X_train, y_train)
        self.encoder = self.clf
        self.fc1 = nn.Linear(16384, 256)
        self.drop = nn.Dropout(p=0.25)
        self.fc2 = nn.Linear(256, 64)

    def forward(self, x, encode=False):
        """Returns the embeddings of the input data if encode is True, else returns the output of the model.

        Args:
            x: torch.Tensor
            encode: bool, optional (default=False)

        Returns:
            x: numpy.ndarray if encode is True, else torch.Tensor
        """
        if encode:
            x = self.clf.predict_embeds(x)
            return torch.Tensor.numpy(x)
        x = F.relu(self.fc1(x))
        return F.relu(self.fc2(x))

    def instantiate_tabpfn(self, X_train, y_train):
        """Instantiates the TabPFNClassifier model, whereas the .forward method is overwritten to return the embeddings
        in case encode is set to True.

        Args:
            X_train: numpy.ndarray
            y_train: numpy.ndarray

        Returns:
            None
        """

        self.num_samples = X_train.shape[0]
        self.clf = TabPFNClassifier(device="cpu", N_ensemble_configurations=32)
        model = self.clf.model[2]
        ENCODER = model.encoder
        N_OUT = model.n_out
        NINP = model.ninp
        NHEAD = 4
        NHID = model.nhid
        NLAYERS = model.transformer_encoder.num_layers
        Y_ENCODER = model.y_encoder
        tf = TransformerModel(ENCODER, N_OUT, NINP, NHEAD, NHID, NLAYERS, y_encoder=Y_ENCODER)
        tf.transformer_encoder = model.transformer_encoder
        tf.decoder = model.decoder
        self.clf.model = ("inf", "inf", tf)
        self.clf.fit(X_train, y_train)
        self.clf.no_grad = True



class TimeLimitCallback(Callback):
    """TimeLimitCallback

    This class is used for limiting the training time of the TabNet classifier, such that it stops training
    after a certain specified time limit.

    Args:
        time_limit: int

    Attributes:
        time_limit: int
        start_time: float
    """
    def __init__(self, time_limit):
        self.time_limit = time_limit
        self.start_time = None

    def on_epoch_begin(self, epoch, logs=None):
        if self.start_time is None:
            self.start_time = time.time()

    def on_epoch_end(self, epoch, logs=None):
        elapsed_time = time.time() - self.start_time
        if elapsed_time > self.time_limit:
            print(f"Stopping training as the time limit of {self.time_limit} seconds has been reached.")
            return True  # This will stop training



import time
import numpy as np
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC

from alpbench.util.common import fullname


class Ensemble:
    """Ensemble

    This class is used to create an ensemble of estimators. The ensemble can be used to predict the probabilities of the
    ensemble members and the classes of the ensemble members.

    Args:
        estimator: object
        num_estimators: int
        max_neighbors: int (for k nearest neighbors) else None

    Attributes:
        estimator: object (the estimator to construct the ensemble of)
        num_estimators: int (the number of estimators in the ensemble)
        max_neighbors: int  (for k nearest neighbors)
        random_states: list (random states for the ensemble members)
        estimators_: list   (list containing the ensemble members)
        learner_fqn: str    (fully qualified name of the estimator)
    """

    def __init__(self, estimator, num_estimators, max_neighbors=None):
        self.estimator = estimator
        self.num_estimators = num_estimators
        self.max_neighbors = max_neighbors
        self.random_states = [np.random.randint(0, 1000) for _ in range(num_estimators)]
        self.estimators_ = None
        self.learner_fqn = fullname(self.estimator)
        self.init()

    def init(self):
        """Initializes the ensemble members.

        Returns:
            None
        """
        self.estimators_ = []
        if self.learner_fqn == "tabpfn.scripts.transformer_prediction_interface.TabPFNClassifier":
            from alpbench.util.transformer_prediction_interface_ens import TabPFNClassifierEns as TabPFNEns
            self.estimator = TabPFNEns(N_ensemble_configurations=self.num_estimators)
        if self.learner_fqn == "pytorch_tabnet.tab_model.TabNetClassifier":
            for seed in self.random_states:
                from alpbench.util.pytorch_tabnet.tab_model import TabNetClassifier
                self.estimators_.append(TabNetClassifier(seed=seed, verbose=0))
        if self.learner_fqn == "catboost.core.CatBoostClassifier":
            self.num_estimators = self.estimator.tree_count_
        if self.learner_fqn == "xgboost.sklearn.XGBClassifier":
            self.num_estimators = self.estimator.n_estimators
        if self.learner_fqn == "sklearn.ensemble._forest.RandomForestClassifier":
            self.estimators_ = self.estimator.estimators_
        if self.learner_fqn == "sklearn.svm._classes.SVC":
            for seed in np.linspace(1, 20, self.num_estimators):
                self.estimators_.append(SVC(kernel="rbf", probability=True, C=seed))
        if self.learner_fqn == "sklearn.neighbors._classification.KNeighborsClassifier":
            num_neighbors = np.random.choice(
                np.linspace(1, self.max_neighbors, self.max_neighbors, dtype=int), self.num_estimators
            )
            for seed in num_neighbors:
                self.estimators_.append(KNeighborsClassifier(n_neighbors=seed))
        else:
            self.estimators_ = [self.estimator for _ in range(self.num_estimators)]

    def fit(self, X: np.ndarray, y: np.ndarray):
        """
        Fits the ensemble and sets the attributes of the class.

        Args
        ----------
        X : array-like of shape (n_samples, n_features)
            The training input samples.
        y : array-like of shape (n_samples,)
        The target values (class labels) as integers.

        Returns
        -------
        None
        """
        if self.learner_fqn in [
            "sklearn.ensemble._forest.RandomForestClassifier",
            "xgboost.sklearn.XGBClassifier",
            "catboost.core.CatBoostClassifier",
            "tabpfn.scripts.transformer_prediction_interface.TabPFNClassifier",
        ]:
            if self.learner_fqn == "tabpfn.scripts.transformer_prediction_interface.TabPFNClassifier" and len(y) > 1000:
                ids = np.random.choice(len(y), 1000)
                self.estimator.fit(X[ids], y[ids])
            else:
                self.estimator.fit(X, y)
        else:
            for estimator in self.estimators_:
                if self.learner_fqn == "pytorch_tabnet.tab_model.TabNetClassifier":
                    from alpbench.util.TorchUtil import TimeLimitCallback
                    estimator.fit(X, y, callbacks=[TimeLimitCallback(60)])
                else:
                    estimator.fit(X, y)
        self.n_classes_ = len(np.unique(y))

    def predict_proba(self, X: np.ndarray, alpha: float = None) -> np.ndarray:
        """
        Predicts the probabilities of the ensemble members.

        Args
        ----------
        X : array-like of shape (n_samples, n_features)
            The training input samples.
            alpha : float, optional (default=None)
            The threshold for the normalized likelihoods of the ensemble members.

        Returns
        -------
        preds : predicted probabilities, array-like of shape (n_samples, n_classes, n_estimators)

        """
        if self.learner_fqn == "tabpfn.scripts.transformer_prediction_interface.TabPFNClassifier":
            return self.estimator.predict_proba(X).transpose(0, 2, 1)
        else:
            preds = np.empty((X.shape[0], self.n_classes_, self.num_estimators))
            for i in range(self.num_estimators):
                if self.learner_fqn == "xgboost.sklearn.XGBClassifier":
                    preds[:, :, i] = self.estimator.predict_proba(X, iteration_range=(i, i + 1))
                elif self.learner_fqn == "catboost.core.CatBoostClassifier":
                    preds[:, :, i] = self.estimator.predict_proba(X, ntree_start=i, ntree_end=i + 1)
                elif self.learner_fqn == "sklearn.ensemble._forest.RandomForestClassifier":
                    preds[:, :, i] = self.estimator.estimators_[i].predict_proba(X)
                else:
                    preds[:, :, i] = self.estimators_[i].predict_proba(X)

            return preds

    def predict(self, X: np.ndarray, alpha: float = None) -> np.ndarray:
        """
        Predicts the classes of the ensemble members.

        Args
        ----------
        X : array-like of shape (n_samples, n_features)
            The training input samples.
            alpha : float, optional (default=None)
            The threshold for the normalized likelihoods of the ensemble members.

        Returns
        -------
        preds : predicted classes, array-like of shape (n_samples, n_estimators)
        """

        probas = self.predict_proba(X, alpha)
        return np.argmax(probas.mean(axis=-1), axis=1)



import pytest
from sklearn.ensemble import RandomForestClassifier as RF

from alpbench.benchmark.Observer import Observer
from alpbench.evaluation.experimenter.LogTableObserver import LogTableObserver, SparseLogTableObserver
from alpbench.pipeline.ActiveLearningPipeline import ActiveLearningPipeline
from alpbench.pipeline.Initializer import RandomInitializer
from alpbench.pipeline.Oracle import Oracle
from alpbench.pipeline.QueryStrategy import RandomQueryStrategy


class MockupLogTableObserver(LogTableObserver):
    def __init__(self):
        pass

    def observe_data(self, iteration, X_u_selected, y_u_selected, X_l_aug, y_l_aug, X_u_red):
        pass

    def observe_model(self, iteration, model):
        pass


class MockupSparseLogTableObserver(SparseLogTableObserver):
    def __init__(self):
        pass

    def observe_data(self, iteration, X_u_selected, y_u_selected, X_l_aug, y_l_aug, X_u_red):
        pass

    def observe_model(self, iteration, model):
        pass

    def log_data(self, dict):
        pass

    def log_model(self, dict):
        pass


class DummyObserver(Observer):
    def __init__(self):
        self.data_logs = False
        self.model_logs = False

    def observe_data(self, iteration, X_u_selected, y_u_selected, X_l_aug, y_l_aug, X_u_red):
        self.data_logs = True

    def observe_model(self, iteration, model):
        self.model_logs = True


@pytest.mark.usefixtures("scenario")
def test_pipeline_executions(scenario):
    print("get data split")
    X_l, y_l, X_u, y_u, X_test, y_test = scenario.get_data_split()

    print("setup learner")
    learner = RF(n_estimators=10)
    query_strategy = RandomQueryStrategy(42)

    ALP = ActiveLearningPipeline(
        learner=learner,
        query_strategy=query_strategy,
        init_budget=10,
        num_iterations=1,
        num_queries_per_iteration=10,
    )

    oracle = Oracle(X_u, y_u)
    ALP.active_fit(X_l, y_l, X_u, oracle)
    y_test_hat = ALP.predict(X_test)
    assert len(y_test_hat) == len(y_test)


@pytest.mark.usefixtures("scenario")
def test_pipeline_execution_with_initializer(scenario):
    print("get data split")
    X_l, y_l, X_u, y_u, X_test, y_test = scenario.get_data_split()

    print("setup learner")
    initializer = RandomInitializer(42)
    learner = RF(n_estimators=10)
    query_strategy = RandomQueryStrategy(42)

    ALP = ActiveLearningPipeline(
        initializer=initializer,
        learner=learner,
        query_strategy=query_strategy,
        init_budget=10,
        num_iterations=1,
        num_queries_per_iteration=10,
    )

    oracle = Oracle(X_u, y_u)
    ALP.active_fit(X_l, y_l, X_u, oracle)
    y_test_hat = ALP.predict(X_test)
    assert len(y_test_hat) == len(y_test)


@pytest.mark.usefixtures("scenario")
def test_pipeline_execution_with_observers(scenario):
    X_l, y_l, X_u, y_u, X_test, y_test = scenario.get_data_split()
    initializers = [None, RandomInitializer(42)]

    for initializer in initializers:
        learner = RF(n_estimators=10)
        query_strategy = RandomQueryStrategy(42)

        ALP = ActiveLearningPipeline(
            initializer=initializer,
            learner=learner,
            query_strategy=query_strategy,
            observer_list=[MockupLogTableObserver(), MockupSparseLogTableObserver()],
            init_budget=10,
            num_iterations=1,
            num_queries_per_iteration=10,
        )

        oracle = Oracle(X_u, y_u)
        ALP.active_fit(X_l, y_l, X_u, oracle)
        y_test_hat = ALP.predict(X_test)
        assert len(y_test_hat) == len(y_test)


@pytest.mark.usefixtures("scenario")
def test_fewer_instances_than_budget(scenario):
    X_l, y_l, X_u, y_u, X_test, y_test = scenario.get_data_split()
    learner = RF(n_estimators=10)
    query_strategy = RandomQueryStrategy(42)

    ALP = ActiveLearningPipeline(
        learner=learner,
        query_strategy=query_strategy,
        init_budget=10,
        num_iterations=2,
        num_queries_per_iteration=10,
    )

    oracle = Oracle(X_u, y_u)
    ALP.active_fit(X_l, y_l, X_u[0:12], oracle)
    y_test_hat = ALP.predict(X_test)
    assert len(y_test_hat) == len(y_test)



import json
import os
import numpy as np
import pandas as pd
import pandas as pd
from py_experimenter.experimenter import PyExperimenter
from scipy.stats import ttest_ind_from_stats
from scipy.stats import ttest_ind as tt
from get_results import StudyData
from plot_functions import HeatMapPlot, BudgetPerformancePlot, WinMatrixPlot


# DEFINE METRIC, DATABASE NAME, GENERATE PANDAS DATAFRAME AND SAVE IT
def generate_aubc_df():
    grid = {}
    metric = ["test_accuracy"]

    db_name = "alpbench_ecai"
    setting_string = "small"
    db_config_file = "../../evaluation/experimenter/config/db_conf.yml"
    exp_scenario_file = "../../evaluation/experimenter/config/experiments.yml"
    grid["db_name"] = db_name
    grid["setting_name"] = setting_string
    source = db_config_file, exp_scenario_file

    studyData = StudyData(source=source, grid=grid)
    end_df = studyData.get_aubc_df()
    # check if path exists
    directory_path = "DATAFRAMES/" + setting_string + "/"
    if not os.path.exists(directory_path):
        print("does not exist")
        os.makedirs(directory_path)
    end_df.to_csv(directory_path + db_name +"_"+ setting_string + "_dynamic.csv")
    return end_df

# GENERATE BUDGET PERFORMANCE PLOT
def generate_budget_performance_plots():
    dataframe = pd.read_csv("DATAFRAMES/small/alpbench_ecai_small_dynamic.csv")

    openmlids = dataframe['openml_id'].unique()
    learners = dataframe['learner_name'].unique()

    collect_results = {}

    for learner in learners[:]:
        collect_results[learner] = {}
        for openmlid in openmlids[:]:
            perfPlot = BudgetPerformancePlot(dataframe, openmlid, learner, "test_accuracy")
            perfPlot.generate_plot_data()

            collect_results[learner][openmlid] = perfPlot.plot_data
            perfPlot.show()

    # save pkl file
    import pickle
    with open('saved_dictionary.pkl', 'wb') as f:
        pickle.dump(collect_results, f)

#generate_aubc_df()
generate_budget_performance_plots()


# GENERATE HEATMAPS

"""
dataframe = pd.read_csv("DATAFRAMES/small/alpbench_ecai_small_dynamic.csv")
learners = dataframe['learner_name'].unique()
filter = ["binary", "multi", "all"]
significance = [True, False]
for learner in learners[:1]:
    for f in filter[:1]:
        for s in significance[:]:
            heatMap = WinMatrixPlot(dataframe, learner, filter_ids=f, take_statistical_insignificant=s)
"""


# GENERATE WIN MATRICES


import json

import numpy as np

from alpbench.benchmark.Observer import StatisticalPerformanceObserver
from alpbench.evaluation.experimenter.LogTableObserver import LogTableObserver, SparseLogTableObserver
from alpbench.util.common import fullname


class NewObserver(StatisticalPerformanceObserver):
    def __init__(self, X_test, y_test):
        super().__init__(X_test, y_test)
        self.X_test = X_test
        self.y_test = y_test

    def observe_model(self, iteration, model):
        return super().compute_model_performances(iteration, model)["test_accuracy"]

    def observe_data(self, iteration, X_u_selected, y_u_selected, X_l_aug, y_l_aug, X_u_red):
        pass


class ActiveLearningPipeline:
    """Active Learning Pipeline

    This class implements an active learning pipeline that can be used to train a model on a given labeled dataset and
    iteratively query an oracle for additional labels to improve the model's performance. The pipeline consists of an
    initializer, a learner, a query strategy, and an observer. The initializer is used to select data points from
    unlabeled data to be labeled by the oracle in the first iteration. The learner is the model that is trained on
    labeled data. The query strategy is used to select data points from the unlabeled dataset to be labeled by the
    oracle in subsequent iterations. The observer is used to log the data and model at each iteration.

    Args:
        learner: The learner that is trained on the labeled data.
        query_strategy: The query strategy used to select data points from the unlabeled dataset to be labeled by
            the oracle.
        initializer: The initializer used to select data points from the unlabeled dataset to be labeled by the oracle
            in the first iterations.
        observer_list: A list of observers that are used to log the data and model at each iteration.
        init_budget: The budget for sampling data points with the initialization strategy.
        num_iterations: The number of active learning rounds to carry out alternating between learning and querying the
            oracle.
        num_queries_per_iteration: The number of data points to select in every active learning iteration to be labeled
            by the oracle.

    Attributes:
        initializer: The initializer used to select data points from the unlabeled dataset to be labeled by the oracle
            in the first iteration.
        learner: The learner that is trained on the labeled data.
        query_strategy: The query strategy used to select data points from the unlabeled dataset to be labeled by
            the oracle.
        observer_list: A list of observers that are used to log the data and model at each iteration.
        init_budget: The budget for sampling data points with the initialization strategy.
        num_iterations: The number of active learning rounds to carry out alternating between learning and querying the
            oracle.
        num_queries_per_iteration: The number of data points to select in every active learning iteration to be labeled
            by the oracle.
    """

    def __init__(
        self,
        learner,
        query_strategy,
        initializer=None,
        observer_list: list() = None,
        init_budget: int = None,
        num_iterations=10,
        num_queries_per_iteration=10,
    ):
        self.initializer = initializer
        self.learner = learner
        self.query_strategy = query_strategy
        self.observer_list = observer_list

        # the budget for sampling data points with the initialization strategy
        self.init_budget = init_budget
        # the number of active learning rounds to carry out alternating between learning and querying the oracle
        self.num_iterations = num_iterations
        # the number of data points to select in every active learning iteration to be labeled by the oracle
        self.num_queries_per_iteration = num_queries_per_iteration

    def active_fit(self, X_l, y_l, X_u, oracle):
        """
        Fits the active learning pipeline. This involves fitting the learner on the current labeled data,
        querying instances from the pool of unlabeled data, forwarding them to the oracle for labeling,
        updating the labeled data with the queried instances and their corresponding labels and repeating the process
        for the given number of iterations. The performances and labeling statistics are logged at each iteration.

        Parameters:
            X_l (np.ndarray): The labeled data.
            y_l (np.ndarray): The labels of the labeled data.
            X_u (np.ndarray): The unlabeled data.
            oracle (Oracle): The oracle used to query the true labels of the data.

        Returns:
            None
        """

        # select data points from X_u to sample additional data points for initialization (i.e., uninformed) and remove
        # the sampled data points from the unlabeled dataset
        idx_available = np.arange(0, len(X_u))
        idx_queried = np.array([])

        X_u_red = X_u
        X_l_aug = X_l
        y_l_aug = y_l

        observer_data = {}
        observer_model = {}

        all_data_used = False

        if self.initializer is not None and self.init_budget is not None:
            idx_init = self.initializer.sample(X_u, self.init_budget)
            # find the index of the sampled indices
            idx_mapped = np.array([np.where(idx_available == value)[0][0] for value in idx_init])
            # update the list of already queried indices
            idx_queried = np.concatenate((idx_queried, idx_init))
            # remove the queried indices from the list of available indices
            np.delete(idx_available, idx_mapped)

            X_u_red = X_u[idx_available]
            X_u_sel = X_u[idx_mapped]

            # label data points via the oracle
            y_u_sel = oracle.query(idx_mapped)

            # augment the given labeled data set by the data points selected for initialization
            X_l_aug = np.concatenate((X_l_aug, X_u_sel))
            y_l_aug = np.concatenate((y_l_aug, y_u_sel))

            if self.observer_list is not None:
                for o in self.observer_list:
                    if isinstance(o, LogTableObserver):
                        o.observe_data(0, X_u_sel, y_u_sel, X_l_aug, y_l_aug, X_u_red)
                    elif isinstance(o, SparseLogTableObserver):
                        observer_data[0] = o.observe_data(0, X_u_sel, y_u_sel, X_l_aug, y_l_aug, X_u_red)

        elif self.observer_list is not None:
            for o in self.observer_list:
                if isinstance(o, LogTableObserver):
                    o.observe_data(0, X_l, y_l, X_l_aug, y_l_aug, X_u_red)
                elif isinstance(o, SparseLogTableObserver):
                    observer_data[0] = o.observe_data(0, X_l, y_l, X_l_aug, y_l_aug, X_u_red)

        # fit the initial model
        learner_fqn = fullname(self.learner)
        if learner_fqn == "tabpfn.scripts.transformer_prediction_interface.TabPFNClassifier" and len(y_l_aug) > 1000:
            ids = np.random.choice(len(y_l_aug), 1000)
            self.learner.fit(X_l_aug[ids], y_l_aug[ids])
        else:
            self.learner.fit(X_l_aug, y_l_aug)

        assert len(np.unique(y_l_aug)) == len(np.unique(y_l)), "Not all classes are represented in the labeled data"

        # let the observer know about the learned model
        if self.observer_list is not None:
            for o in self.observer_list:
                if isinstance(o, LogTableObserver):
                    o.observe_model(0, self.learner)
                elif isinstance(o, SparseLogTableObserver):
                    observer_model[0] = o.observe_model(iteration=0, model=self.learner)

        for i in range(1, self.num_iterations + 1):

            if self.num_queries_per_iteration > len(idx_available):
                idx_query_orig = idx_available
                all_data_used = True

            else:
                # ask query strategy for samples
                idx_query = self.query_strategy.sample(
                    learner=self.learner,
                    X_l=X_l_aug,
                    y_l=y_l_aug,
                    X_u=X_u_red,
                    num_queries=self.num_queries_per_iteration,
                )
                # get the original indices for X_u
                idx_query_orig = idx_available[idx_query]

            idx_mapping = np.array([np.where(idx_available == v)[0][0] for v in idx_query_orig])

            # delete the selected indices from the available list of indices
            idx_available = np.delete(idx_available, idx_mapping)

            # if dimension > 1
            if len(idx_query_orig.shape) > 1:
                idx_query_orig = idx_query_orig.flatten()

            idx_queried = np.concatenate((idx_queried, idx_query_orig))

            # print("IDX queried: ", idx_queried)

            X_u_red = X_u[idx_available]
            X_u_sel = X_u[idx_query_orig]

            # query oracle for ground truth labels
            y_u_sel = oracle.query(idx_query_orig)

            # augment the labeled dataset
            X_l_aug = np.concatenate([X_l_aug, X_u_sel])
            y_l_aug = np.concatenate([y_l_aug, y_u_sel])

            # let the observer see the change in the data for this iteration
            if self.observer_list is not None:
                for o in self.observer_list:
                    if isinstance(o, LogTableObserver):
                        o.observe_data(i, X_u_sel, y_u_sel, X_l_aug, y_l_aug, X_u_red)
                    elif isinstance(o, SparseLogTableObserver):
                        observer_data[i] = o.observe_data(i, X_u_sel, y_u_sel, X_l_aug, y_l_aug, X_u_red)

            # fit the initial model
            learner_fqn = fullname(self.learner)
            if (
                learner_fqn == "tabpfn.scripts.transformer_prediction_interface.TabPFNClassifier"
                and len(y_l_aug) > 1000
            ):
                ids = np.random.choice(len(y_l_aug), 1000)
                self.learner.fit(X_l_aug[ids], y_l_aug[ids])
            elif learner_fqn == "pytorch_tabnet.tab_model.TabNetClassifier":
                from alpbench.util.TorchUtil import TimeLimitCallback

                self.learner.fit(X_l_aug, y_l_aug, callbacks=[TimeLimitCallback(180)])
            else:
                self.learner.fit(X_l_aug, y_l_aug)

            # let the observer know about the learned model
            if self.observer_list is not None:
                for o in self.observer_list:
                    if isinstance(o, LogTableObserver):
                        o.observe_model(i, self.learner)
                    elif isinstance(o, SparseLogTableObserver):
                        observer_model[i] = o.observe_model(i, self.learner)

            if all_data_used:
                break

        # finales logging
        if self.observer_list is not None:
            for o in self.observer_list:
                if isinstance(o, SparseLogTableObserver):
                    o.log_data({"data_dict": json.dumps(observer_data)})
                    o.log_model({"model_dict": json.dumps(observer_model)})

    def predict(self, X):
        """Lets the learner predict on the given data.

        Parameters:
            X (np.ndarray): The data to predict on.

        Returns:
            np.ndarray: The predictions of the learner on the given data.
        """
        return self.learner.predict(X)



import numpy as np
from py_experimenter.exceptions import DatabaseConnectionError
from py_experimenter.experimenter import PyExperimenter, ResultProcessor
import types
from alpbench.benchmark.BenchmarkConnector import DataFileBenchmarkConnector
from alpbench.evaluation.experimenter.DefaultSetup import ensure_default_setup
from alpbench.evaluation.experimenter.LogTableObserver import LogTableObserver, SparseLogTableObserver
from alpbench.pipeline.ActiveLearningPipeline import ActiveLearningPipeline
from alpbench.pipeline.Oracle import Oracle
import sqlite3
import pandas as pd
import os


# loads parameters from the grid and runs each active learning pipeline on every active learning problem
class ExperimentRunner:

    def __init__(self):
        pass

    def run_experiment(self, parameters: dict, result_processor: ResultProcessor, custom_config: dict):

        dbbc = DataFileBenchmarkConnector()

        connector: DataFileBenchmarkConnector = dbbc

        OPENML_ID = int(parameters["openml_id"])
        SETTING_NAME = parameters["setting_name"]
        TEST_SPLIT_SEED = int(parameters["test_split_seed"])
        TRAIN_SPLIT_SEED = int(parameters["train_split_seed"])
        SEED = int(parameters["seed"])

        setting = connector.load_setting_by_name(SETTING_NAME)
        scenario = connector.load_or_create_scenario(
            openml_id=OPENML_ID,
            test_split_seed=TEST_SPLIT_SEED,
            train_split_seed=TRAIN_SPLIT_SEED,
            seed=SEED,
            setting_id=setting.get_setting_id(),
        )

        X_l, y_l, X_u, y_u, X_test, y_test = scenario.get_data_split()

        QUERY_STRATEGY = connector.load_query_strategy_by_name(parameters["query_strategy_name"])

        LEARNER = connector.load_learner_by_name(parameters["learner_name"])

        OBSERVER = [SparseLogTableObserver(result_processor, X_test, y_test)]

        ALP = ActiveLearningPipeline(
            learner=LEARNER,
            query_strategy=QUERY_STRATEGY,
            observer_list=OBSERVER,
            num_iterations=setting.get_number_of_iterations(),
            num_queries_per_iteration=setting.get_number_of_queries(),
        )

        oracle = Oracle(X_u, y_u)
        ALP.active_fit(X_l, y_l, X_u, oracle)



def run(db_spec, run_setup=False, reset_experiments=False, setting_combinations=None, parameters=None):


    experimenter = PyExperimenter(experiment_configuration_file_path=db_spec)

    if run_setup:
        benchmark_connector = DataFileBenchmarkConnector()
        ensure_default_setup(dbbc=benchmark_connector)

        benchmark_connector.cleanup()

        setting_combinations = []
        setting_combinations += [{"setting_name": "small"}]

        if reset_experiments:
            experimenter.reset_experiments("running", "failed")

        else:
            experimenter.fill_table_from_combination(
                parameters=parameters,
                fixed_parameter_combinations=setting_combinations,
            )

    else:
        er = ExperimentRunner()
        experimenter.execute(er.run_experiment, -1)


import pytest
from sklearn.ensemble import RandomForestClassifier as RF

from alpbench.pipeline.ActiveLearningPipeline import ActiveLearningPipeline
from alpbench.pipeline.Oracle import Oracle
from alpbench.pipeline.QueryStrategy import (
    BALDQueryStrategy,
    ClusterMarginQueryStrategy,
    CoreSetQueryStrategy,
    DiscriminativeQueryStrategy,
    EntropyQueryStrategy,
    FalcunQueryStrategy,
    KMeansQueryStrategy,
    LeastConfidentQueryStrategy,
    MarginQueryStrategy,
    MaxEntropyQueryStrategy,
    MinMarginQueryStrategy,
    PowerBALDQueryStrategy,
    PowerMarginQueryStrategy,
    QBCVarianceRatioQueryStrategy,
    QueryByCommitteeEntropyQueryStrategy,
    QueryByCommitteeKLQueryStrategy,
    RandomMarginQueryStrategy,
    RandomQueryStrategy,
    TypicalClusterQueryStrategy,
    WeightedClusterQueryStrategy,
)

default_query_strategies = {
    "random": RandomQueryStrategy(42),
    "random_margin": RandomMarginQueryStrategy(42),
    "cluster_margin": ClusterMarginQueryStrategy(42),
    "core_set": CoreSetQueryStrategy(42),
    "entropy": EntropyQueryStrategy(42),
    "falcun": FalcunQueryStrategy(42),
    "margin": MarginQueryStrategy(42),
    "max_entropy": MaxEntropyQueryStrategy(42, 10),
    "least_confident": LeastConfidentQueryStrategy(42),
    "kmeans": KMeansQueryStrategy(42),
    "discrim": DiscriminativeQueryStrategy(42),
    "qbc_entropy": QueryByCommitteeEntropyQueryStrategy(42, 10),
    "qbc_kl": QueryByCommitteeKLQueryStrategy(42, 10),
    "bald": BALDQueryStrategy(42, 10),
    "power_margin": PowerMarginQueryStrategy(42),
    "min_margin": MinMarginQueryStrategy(42, 10),
    "typ_cluster": TypicalClusterQueryStrategy(42),
    "weighted_cluster": WeightedClusterQueryStrategy(42),
    "qbc_variance_ratio": QBCVarianceRatioQueryStrategy(42, 10),
    "power_bald": PowerBALDQueryStrategy(42, 10),
}


@pytest.mark.usefixtures("scenario")
def test_pipeline_executions(scenario):
    print("get data split")
    X_l, y_l, X_u, y_u, X_test, y_test = scenario.get_data_split()
    print("defined sampling strategies")
    query_strategies = [value for key, value in default_query_strategies.items()]
    print("setup learner")
    learner = RF(n_estimators=10)

    print("test sampling strategies")
    for query_strategy in query_strategies:
        print("current sampling strategy: ", query_strategy)
        ALP = ActiveLearningPipeline(
            learner=learner,
            query_strategy=query_strategy,
            init_budget=10,
            num_iterations=1,
            num_queries_per_iteration=10,
        )

        oracle = Oracle(X_u, y_u)
        ALP.active_fit(X_l, y_l, X_u, oracle)



from alpbench.benchmark.BenchmarkConnector import BenchmarkConnector
from alpbench.benchmark.Observer import Observer
from alpbench.pipeline.ActiveLearningPipeline import ActiveLearningPipeline
from alpbench.pipeline.Oracle import Oracle


class ALPEvaluator:
    """ALPEvaluator

    This class is used to evaluate an active learning pipeline (alpbench) on a specific benchmark setting. This involves
    fitting the alpbench on the training data and then evaluating it on the test data.

    Args:
        benchmark_connector (BenchmarkConnector): The benchmark connector to use.
        setting_name (str): The name of the setting to use.
        learner_name (str): The name of the learner to use.
        query_strategy_name (str): The name of the query strategy to use.
        openml_id (int): The openml id of the benchmark scenario to use.
        test_split_seed (int): The seed to use for splitting the test data.
        train_split_seed (int): The seed to use for splitting the train data.
        seed (int): The seed to use for the random number generator.

    Attributes:
        benchmark_connector (BenchmarkConnector): The benchmark connector to use.
        observer_list (list): The list of observers to use.
        setting (Setting): The setting to use.
        scenario (Scenario): The scenario to use.
        setting_name (str): The name of the setting to use.
        learner_name (str): The name of the learner to use.
        query_strategy_name (str): The name of the query strategy to use.
        openml_id (int): The openml id of the benchmark scenario to use.
        test_split_seed (int): The seed to use for splitting the test data.
        train_split_seed (int): The seed to use for splitting the train data.
        seed (int): The seed to use for the random number generator.

    """

    def __init__(
        self,
        benchmark_connector: BenchmarkConnector,
        setting_name=None,
        learner_name=None,
        query_strategy_name=None,
        openml_id=None,
        test_split_seed=0,
        train_split_seed=0,
        seed=0,
    ):
        self.benchmark_connector = benchmark_connector
        self.observer_list = []
        self.setting = None
        self.scenario = None

        # setting specifications
        self.setting_name = setting_name

        # scenario specifications
        self.openml_id = openml_id
        self.test_split_seed = test_split_seed
        self.train_split_seed = train_split_seed
        self.seed = seed

        # pipeline specifications
        self.learner_name = learner_name
        self.query_strategy_name = query_strategy_name

    def with_setting(self, setting_name: str):
        self.setting_name = setting_name
        return self

    def with_learner(self, learner_name: str):
        self.learner_name = learner_name
        return self

    def with_query_strategy(self, query_strategy_name: str):
        self.query_strategy_name = query_strategy_name
        return self

    def with_test_split_seed(self, test_split_seed: int):
        self.test_split_seed = test_split_seed
        return self

    def with_train_split_seed(self, train_split_seed: int):
        self.train_split_seed = train_split_seed
        return self

    def with_openml_id(self, openml_id):
        self.openml_id = openml_id
        return self

    def with_observer(self, observer: Observer):
        self.observer_list += [observer]
        return self

    def _load_setting_and_scenario(self):
        assert self.openml_id is not None and self.setting_name is not None
        self.setting = self.benchmark_connector.load_setting_by_name(self.setting_name)
        self.scenario = self.benchmark_connector.load_or_create_scenario(
            openml_id=self.openml_id,
            test_split_seed=self.test_split_seed,
            train_split_seed=self.train_split_seed,
            seed=self.seed,
            setting_id=self.setting.get_setting_id(),
        )

    def get_test_data(self):
        self._load_setting_and_scenario()
        return self.scenario.get_test_data()

    def fit(self):
        self._load_setting_and_scenario()

        X_l, y_l, X_u, y_u, X_test, y_test = self.scenario.get_data_split()
        learner = self.benchmark_connector.load_learner_by_name(self.learner_name)
        query_strategy = self.benchmark_connector.load_query_strategy_by_name(self.query_strategy_name)

        oracle = Oracle(X_u=X_u, y_u=y_u)
        alp = ActiveLearningPipeline(
            learner=learner,
            query_strategy=query_strategy,
            num_iterations=self.setting.get_number_of_iterations(),
            observer_list=self.observer_list,
            num_queries_per_iteration=self.setting.get_number_of_queries(),
        )
        alp.active_fit(X_l=X_l, y_l=y_l, X_u=X_u, oracle=oracle)

        return alp



from abc import ABC, abstractmethod

import numpy as np
from sklearn.metrics import accuracy_score, f1_score, log_loss, precision_score, recall_score, roc_auc_score


class Observer(ABC):
    """Observer

    Abstract class for the observer of the active learning process. The observer is responsible for keeping track of
    the indices of the selected data, the labeled data, and the unlabeled data, as well as the model performance on the
    test data in each iteration..
    """

    def __init__(self):
        return

    @abstractmethod
    def observe_data(self, iteration, X_u_selected, y_u_selected, X_l_aug, y_l_aug, X_u_red):
        """
        Abstract method to observe the data in each iteration.
        """
        pass

    @abstractmethod
    def observe_model(self, iteration, model):
        """
        Abstract method to observe the model performances in each iteration.
        """
        pass


class StatisticalPerformanceObserver(Observer, ABC):
    """StatisticalPerformanceObserver

    Observer class for statistical performance evaluation of the active learning process such as the distribution of
    the labeled and selected data, and the model performance on the test data.

    Args:
        X_test (np.array): test data
        y_test (np.array): test labels

    Attributes:
        X_test (np.array): test data
        y_test (np.array): test labels
        precision (int): precision of the floating point numbers
    """

    precision = 8

    def __init__(self, X_test, y_test):
        super().__init__()
        self.X_test = X_test
        self.y_test = y_test

    def compute_labeling_statistics(self, iteration, X_u_selected, y_u_selected, X_l_aug, y_l_aug, X_u_red):
        """
        Computes the distribution of the data, which involves the ids of the selected data in each iteration, the
        overall labeled and unlabeled data.

        Parameters:
            iteration (int): iteration number
            X_u_selected (np.array): selected data
            y_u_selected (np.array): selected labels
            X_l_aug (np.array): labeled data
            y_l_aug (np.array): labeled labels
            X_u_red (np.array): unlabeled data

        Returns:
            eval_scores (dict): dictionary with the distribution of the labeled and selected data

        """
        if len(y_u_selected) > 1:
            unique_values, counts = np.unique(np.array(y_u_selected), return_counts=True)
            selected_dist = dict(zip(unique_values, (counts / len(y_u_selected))))
            for k in selected_dist.keys():
                selected_dist[k] = round(selected_dist[k], StatisticalPerformanceObserver.precision)
        else:
            selected_dist = {str(y_u_selected): 1}

        unique_values, counts = np.unique(y_l_aug, return_counts=True)
        labeled_dist = dict(zip(unique_values, counts / len(y_l_aug)))
        for k in labeled_dist.keys():
            labeled_dist[k] = round(labeled_dist[k], StatisticalPerformanceObserver.precision)

        eval_scores = {
            "iteration": iteration,
            "len_X_sel": len(X_u_selected),
            "len_X_l": len(X_l_aug),
            "len_X_u": len(X_u_red),
            "y_sel_dist": str(selected_dist),
            "y_l_dist": str(labeled_dist),
        }
        return eval_scores

    def compute_model_performances(self, iteration, model):
        """
        Computes the model performance on the test data, which involves the f1, precision, recall, log loss, accuracy,
        and AUC scores.

        Parameters:
            iteration (int): iteration number
            model (object): trained model

        Returns:
            eval_scores (dict): dictionary with the model performance on the test data
        """
        eval_scores = {"iteration": iteration}
        y_hat = model.predict(self.X_test)
        y_hat_proba = model.predict_proba(self.X_test)

        eval_scores = {
            "iteration": iteration,
            "test_f1": round(
                f1_score(self.y_test, y_hat, average="weighted"), StatisticalPerformanceObserver.precision
            ),
            "test_precision": round(
                precision_score(self.y_test, y_hat, average="weighted"), StatisticalPerformanceObserver.precision
            ),
            "test_recall": round(
                recall_score(self.y_test, y_hat, average="weighted"), StatisticalPerformanceObserver.precision
            ),
            "test_log_loss": round(log_loss(self.y_test, y_hat_proba), StatisticalPerformanceObserver.precision),
            "test_accuracy": round(accuracy_score(self.y_test, y_hat), StatisticalPerformanceObserver.precision),
        }

        if len(np.unique(self.y_test)) == 2:
            eval_scores["test_auc"] = round(
                roc_auc_score(self.y_test, y_hat_proba[:, 1]), StatisticalPerformanceObserver.precision
            )
        else:
            eval_scores["test_auc"] = round(
                roc_auc_score(self.y_test, y_hat_proba, multi_class="ovr"), StatisticalPerformanceObserver.precision
            )
        return eval_scores


class PrintObserver(StatisticalPerformanceObserver):
    """PrintObserver

    Observer class to get the statistical performance evaluation of the active learning process such as the
    distribution of the labeled and selected data, and the model performance on the test data.

    Args:
        X_test (np.array): test data
        y_test (np.array): test labels

    Attributes:
        X_test (np.array): test data
        y_test (np.array): test labels
    """

    def __init__(self, X_test, y_test):
        super().__init__(X_test, y_test)
        self.X_test = X_test
        self.y_test = y_test

    def observe_data(self, iteration, X_u_selected, y_u_selected, X_l_aug, y_l_aug, X_u_red):
        """
        Observes the data in each iteration.

        Parameters:
            iteration (int): iteration number
            X_u_selected (np.array): selected data
            y_u_selected (np.array): selected labels
            X_l_aug (np.array): labeled data
            y_l_aug (np.array): labels of labeled data
            X_u_red (np.array): unlabeled data
        """
        super().compute_labeling_statistics(iteration, X_u_selected, y_u_selected, X_l_aug, y_l_aug, X_u_red)

    def observe_model(self, iteration, model):
        """
        Observes the model performances in each iteration.

        Parameters:
            iteration (int): iteration number
            model (object): trained model
        """
        super().compute_model_performances(iteration, model)



from abc import ABC, abstractmethod

import numpy as np
from skactiveml.pool import RandomSampling


class Initializer(ABC):
    """Initializer

    This abstract class initially queries instances of the data to costruct the set of labeled data.
    """

    def __init__(self):
        return

    @abstractmethod
    def sample(self, X_u, num_samples):
        """
        Abstract method, that queries the initial instances of the data.
        """
        pass


class RandomInitializer(Initializer):
    """RandomInitializer

    Randomly selects a subset of the unlabeled data to be queried.

    Args:
        seed (int): The seed for the random number generator.
    Attributes:
        qs (class): RandomSampling class from skactiveml.pool.
    """

    def __init__(self, seed):
        super().__init__()
        self.qs = RandomSampling(random_state=seed)

    def sample(self, X_u, num_samples):
        """Queries the initial instances of the data, sampling uniformly at random

        Parameters:
            X_u (np.ndarray): The unlabeled data.
            num_samples (int): The number of instances to query.
        Returns:
            np.ndarray: The indices of the queried instances.
        """
        nan_labels = np.full(len(X_u), np.nan, dtype=float)
        queried_ids = self.qs.query(X=X_u, y=nan_labels, batch_size=num_samples)
        return queried_ids



import numpy as np
from py_experimenter.exceptions import DatabaseConnectionError
from py_experimenter.experimenter import PyExperimenter, ResultProcessor
import types

from alpbench.benchmark.BenchmarkConnector import MySQLBenchmarkConnector, DataFileBenchmarkConnector
from alpbench.benchmark.BenchmarkSuite import (
    OpenMLBenchmarkSuite,
    TabZillaBenchmarkSuite
)
from alpbench.evaluation.experimenter.LogTableObserver import LogTableObserver, SparseLogTableObserver
from alpbench.pipeline.ActiveLearningPipeline import ActiveLearningPipeline
from alpbench.pipeline.Oracle import Oracle

import xgboost as xgb
import catboost as cb

from sklearn.ensemble import (
    ExtraTreesClassifier,
    GradientBoostingClassifier,
    RandomForestClassifier,
)
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB, MultinomialNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.svm import SVC
from tabpfn import TabPFNClassifier
from alpbench.util.pytorch_tabnet.tab_model import TabNetClassifier
from alpbench.benchmark.BenchmarkConnector import MySQLBenchmarkConnector, DataFileBenchmarkConnector

from alpbench.pipeline.QueryStrategy import (
    BALDQueryStrategy,
    ClusterMarginQueryStrategy,
    CoreSetQueryStrategy,
    DiscriminativeQueryStrategy,
    EntropyQueryStrategy,
    EpistemicUncertaintyQueryStrategy,
    FalcunQueryStrategy,
    KMeansQueryStrategy,
    LeastConfidentQueryStrategy,
    MarginQueryStrategy,
    MaxEntropyQueryStrategy,
    MinMarginQueryStrategy,
    MonteCarloEERLogLoss,
    MonteCarloEERMisclassification,
    PowerBALDQueryStrategy,
    PowerMarginQueryStrategy,
    QBCVarianceRatioQueryStrategy,
    QueryByCommitteeEntropyQueryStrategy,
    QueryByCommitteeKLQueryStrategy,
    RandomMarginQueryStrategy,
    RandomQueryStrategy,
    TypicalClusterQueryStrategy,
    WeightedClusterQueryStrategy
)
from alpbench.benchmark.ActiveLearningScenario import ActiveLearningScenario
from alpbench.benchmark.ActiveLearningSetting import ActiveLearningSetting

#exp_learner_sampler_file = "config/exp_learner_sampler.yml"
exp_learner_sampler_file = "config/experiments.yml"
db_config_file = "config/db_conf.yml"

default_learners = {
    "svm_lin": SVC(kernel='linear', probability=True),
    "svm_rbf": SVC(kernel='rbf', probability=True),
    "rf_entropy": RandomForestClassifier(n_estimators=100, max_depth=10, criterion='entropy'),
    "rf_gini": RandomForestClassifier(n_estimators=100, max_depth=10, criterion='gini'),
    "rf_entropy_large": RandomForestClassifier(n_estimators=250, max_depth=10, criterion='entropy'),
    "rf_gini_large": RandomForestClassifier(n_estimators=250, max_depth=10, criterion='gini'),
    "knn_3": KNeighborsClassifier(n_neighbors=3),
    "knn_10": KNeighborsClassifier(n_neighbors=10),
    "log_reg": LogisticRegression(),
    "multinomial_bayes": MultinomialNB(),
    "etc_entropy": ExtraTreesClassifier(n_estimators=100, max_depth=10, criterion='entropy'),
    "etc_gini": ExtraTreesClassifier(n_estimators=100, max_depth=10, criterion='gini'),
    "etc_entropy_large": ExtraTreesClassifier(n_estimators=250, max_depth=10, criterion='entropy'),
    "etc_gini_large": ExtraTreesClassifier(n_estimators=250, max_depth=10, criterion='gini'),
    "naive_bayes": GaussianNB(),
    "mlp": MLPClassifier(),
    "GBT_logloss": GradientBoostingClassifier(n_estimators=100),
    "GBT_exp": GradientBoostingClassifier(n_estimators=100, loss='exponential'),
    "GBT_logloss_large": GradientBoostingClassifier(n_estimators=250),
    "tabpfn": TabPFNClassifier(device='cpu', N_ensemble_configurations=32),
    "xgb": xgb.XGBClassifier(tree_method='hist', max_depth=6, n_estimators=100),
    "catboost": cb.CatBoostClassifier(rsm=0.1, iterations=100, depth=10, verbose=0, boosting_type='Plain', bootstrap_type='Bernoulli', subsample=.75),
    "tabnet": TabNetClassifier(verbose=0)
}

default_query_strategies = {
    "random": RandomQueryStrategy(42),
    "random_margin": RandomMarginQueryStrategy(42),
    "cluster_margin": ClusterMarginQueryStrategy(42),
    "core_set": CoreSetQueryStrategy(42),
    "epistemic": EpistemicUncertaintyQueryStrategy(42),
    "entropy": EntropyQueryStrategy(42),
    "falcun": FalcunQueryStrategy(42),
    "margin": MarginQueryStrategy(42),
    "max_entropy": MaxEntropyQueryStrategy(42, 10),
    "least_confident": LeastConfidentQueryStrategy(42),
    "mc_logloss": MonteCarloEERLogLoss(42),
    "mc_misclass": MonteCarloEERMisclassification(42),
    "kmeans": KMeansQueryStrategy(42),
    "discrim": DiscriminativeQueryStrategy(42),
    "qbc_entropy": QueryByCommitteeEntropyQueryStrategy(42, 10),
    "qbc_kl": QueryByCommitteeKLQueryStrategy(42, 10),
    "bald": BALDQueryStrategy(42, 10),
    "power_margin": PowerMarginQueryStrategy(42),
    "min_margin": MinMarginQueryStrategy(42, 10),
    "typ_cluster": TypicalClusterQueryStrategy(42),
    "weighted_cluster": WeightedClusterQueryStrategy(42),
    "qbc_variance_ratio": QBCVarianceRatioQueryStrategy(42, 10),
    "power_bald": PowerBALDQueryStrategy(42, 10),
}


tabzilla = TabZillaBenchmarkSuite()
tabzilla_ids = tabzilla.get_openml_dataset_ids()
cc18 = OpenMLBenchmarkSuite(99)
cc18_ids = cc18.get_openml_dataset_ids()


class ExperimentRunner:

    def __init__(self, db_credentials, db_name):
        self.db_credentials = db_credentials
        self.db_name = db_name

    def run_experiment(self, parameters: dict, result_processor: ResultProcessor, custom_config: dict):

        dbbc = MySQLBenchmarkConnector(host=self.db_credentials["host"], user=self.db_credentials["user"],
                                       password=self.db_credentials["password"], database=self.db_name, use_ssl=False)
        #try:
        connector: MySQLBenchmarkConnector = dbbc

        OPENML_ID = int(parameters["openml_id"])
        SETTING_NAME = parameters["setting_name"]
        TEST_SPLIT_SEED = int(parameters["test_split_seed"])
        TRAIN_SPLIT_SEED = int(parameters["train_split_seed"])
        SEED = int(parameters["seed"])

        setting = ActiveLearningSetting(setting_id=4, setting_name='small_dynamic', setting_test_size=0.33,
                              setting_train_type='absolute', setting_labeled_train_size=30,
                              number_of_iterations=1, number_of_queries=1, factor=5)


        scenario = connector.load_or_create_scenario(openml_id=OPENML_ID, test_split_seed=TEST_SPLIT_SEED,
                                                     train_split_seed=TRAIN_SPLIT_SEED, seed=SEED,
                                                     setting_id=setting.get_setting_id())

        X_l, y_l, X_u, y_u, X_test, y_test = scenario.get_data_split()

        query_strat_name = parameters["query_strategy_name"]
        learner_name = parameters["learner_name"]


        if learner_name == "tabpfn":
            if X_l.shape[-1]>100 or len(np.unique(y_l))>10:
                dbbc.close()
                return

        LEARNER = default_learners[learner_name]

        task_dependent_query_budget = setting.factor * len(np.unique(np.concatenate(([y_l,y_u,y_test]))))
        setting.number_of_queries = task_dependent_query_budget

        QUERY_STRATEGY = default_query_strategies[query_strat_name]
        LEARNER = default_learners[learner_name]

        #OBSERVER = [LogTableObserver(result_processor, X_test, y_test)]
        OBSERVER = [SparseLogTableObserver(result_processor, X_test, y_test)]


        ALP = ActiveLearningPipeline(learner=LEARNER, query_strategy=QUERY_STRATEGY, observer_list=OBSERVER,
                                     num_iterations=setting.get_number_of_iterations(),
                                     num_queries_per_iteration=setting.get_number_of_queries())

        oracle = Oracle(X_u, y_u)
        ALP.active_fit(X_l, y_l, X_u, oracle)
        #finally:
        print("closing connection")
        dbbc.close()


def run_parallel(variable, default_setup=False, run_setup=False, reset_experiments=False):

    experimenter = PyExperimenter(experiment_configuration_file_path=exp_learner_sampler_file,
                                  database_credential_file_path=db_config_file,
                                  name=variable)

    db_name = experimenter.config.database_configuration.database_name
    db_credentials = experimenter.db_connector._get_database_credentials()

    if run_setup:
        #from DefaultSetup import ensure_default_setup
        #dbbc = MySQLBenchmarkConnector(host=db_credentials["host"], user=db_credentials["user"],
        #                               password=db_credentials["password"], database=db_name, use_ssl=False)
        #ensure_default_setup(dbbc=dbbc)

        setting_combinations = []

        filter_out = [1567,1169,41147,1493]
        setting_combinations += [{'setting_name': 'small', 'openml_id': oid, 'train_split_seed': seed} for oid in cc18_ids if oid not in filter_out for seed in np.arange(20)]


        if reset_experiments:
            experimenter.reset_experiments('running')#, 'failed', 'error')
        else:
            experimenter.fill_table_from_combination(
                parameters={

                    "learner_name": ["rf_entropy",
                                    "tabpfn",
                                     "catboost"
                                     ],
                    "query_strategy_name": ["margin",
                                            "entropy",
                                            "cluster_margin", "typ_cluster",
                                            "random"
                    ],
                    "test_split_seed": np.arange(1),
                    #"train_split_seed": np.arange(1),
                    "seed": np.arange(1)
                },
                fixed_parameter_combinations=setting_combinations)

    else:
        er = ExperimentRunner(db_credentials, db_name)
        experimenter.execute(er.run_experiment, -1)

def main():
    import sys


    variable_job_id = str(sys.argv[1])
    variable_task_id = str(sys.argv[2])


    # concatenate both variables with _
    variable = variable_job_id + "_" + variable_task_id

    run_parallel(variable, run_setup = True, reset_experiments = True)
    #run_parallel(variable)



if __name__ == "__main__":
    main()



import io
import os
import pathlib
import pickle
import random
from pathlib import Path

import numpy as np
import torch
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.preprocessing import LabelEncoder, PowerTransformer, QuantileTransformer, RobustScaler
from sklearn.utils import column_or_1d
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils.validation import check_array, check_is_fitted, check_X_y
from tabpfn.scripts.model_builder import load_model, load_model_only_inference
from tabpfn.utils import NOP, normalize_by_used_features_f, normalize_data, remove_outliers, to_ranking_low_mem
from torch.utils.checkpoint import checkpoint


class CustomUnpickler(pickle.Unpickler):
    def find_class(self, module, name):
        if name == "Manager":
            from settings import Manager

            return Manager
        try:
            return self.find_class_cpu(module, name)
        except:
            return None

    def find_class_cpu(self, module, name):
        if module == "torch.storage" and name == "_load_from_bytes":
            return lambda b: torch.load(io.BytesIO(b), map_location="cpu")
        else:
            return super().find_class(module, name)


def load_model_workflow(i, e, add_name, base_path, device="cpu", eval_addition="", only_inference=True):
    """
    Workflow for loading a model and setting appropriate parameters for diffable hparam tuning.

    :param i:
    :param e:
    :param eval_positions_valid:
    :param add_name:
    :param base_path:
    :param device:
    :param eval_addition:
    :return:
    """

    def get_file(e):
        """
        Returns the different paths of model_file, model_path and results_file
        """
        model_file = f"models_diff/prior_diff_real_checkpoint{add_name}_n_{i}_epoch_{e}.cpkt"
        model_path = os.path.join(base_path, model_file)
        # print('Evaluate ', model_path)
        results_file = os.path.join(
            base_path, f"models_diff/prior_diff_real_results{add_name}_n_{i}_epoch_{e}_{eval_addition}.pkl"
        )
        return model_file, model_path, results_file

    def check_file(e):
        model_file, model_path, results_file = get_file(e)
        if not Path(model_path).is_file():  # or Path(results_file).is_file():
            print("We have to download the TabPFN, as there is no checkpoint at ", model_path)
            print("It has about 100MB, so this might take a moment.")
            import requests

            url = "https://github.com/automl/TabPFN/raw/main/tabpfn/models_diff/prior_diff_real_checkpoint_n_0_epoch_42.cpkt"
            r = requests.get(url, allow_redirects=True)
            os.makedirs(os.path.dirname(model_path), exist_ok=True)
            open(model_path, "wb").write(r.content)
        return model_file, model_path, results_file

    model_file = None
    if e == -1:
        for e_ in range(100, -1, -1):
            model_file_, model_path_, results_file_ = check_file(e_)
            if model_file_ is not None:
                e = e_
                model_file, model_path, results_file = model_file_, model_path_, results_file_
                break
    else:
        model_file, model_path, results_file = check_file(e)

    if model_file is None:
        model_file, model_path, results_file = get_file(e)
        raise Exception("No checkpoint found at " + str(model_path))

    # print(f'Loading {model_file}')
    if only_inference:
        # print('Loading model that can be used for inference only')
        model, c = load_model_only_inference(base_path, model_file, device)
    else:
        # until now also only capable of inference
        model, c = load_model(base_path, model_file, device, eval_positions=[], verbose=False)
    # model, c = load_model(base_path, model_file, device, eval_positions=[], verbose=False)

    return model, c, results_file


class TabPFNClassifier(BaseEstimator, ClassifierMixin):

    models_in_memory = {}

    def __init__(
        self,
        device="cpu",
        base_path=pathlib.Path(__file__).parent.parent.resolve(),
        model_string="",
        N_ensemble_configurations=3,
        no_preprocess_mode=False,
        multiclass_decoder="permutation",
        feature_shift_decoder=True,
        only_inference=True,
        seed=0,
        no_grad=True,
        batch_size_inference=32,
    ):
        """
        Initializes the classifier and loads the model.
        Depending on the arguments, the model is either loaded from memory, from a file, or downloaded from the
        repository if no model is found.

        Can also be used to compute gradients with respect to the inputs X_train and X_test. Therefore no_grad has to be
        set to False and no_preprocessing_mode must be True. Furthermore, X_train and X_test need to be given as
        torch.Tensors and their requires_grad parameter must be set to True.


        :param device: If the model should run on cuda or cpu.
        :param base_path: Base path of the directory, from which the folders like models_diff can be accessed.
        :param model_string: Name of the model. Used first to check if the model is already in memory, and if not,
               tries to load a model with that name from the models_diff directory. It looks for files named as
               follows: "prior_diff_real_checkpoint" + model_string + "_n_0_epoch_e.cpkt", where e can be a number
               between 100 and 0, and is checked in a descending order.
        :param N_ensemble_configurations: The number of ensemble configurations used for the prediction. Thereby the
               accuracy, but also the running time, increases with this number.
        :param no_preprocess_mode: Specifies whether preprocessing is to be performed.
        :param multiclass_decoder: If set to permutation, randomly shifts the classes for each ensemble configuration.
        :param feature_shift_decoder: If set to true shifts the features for each ensemble configuration according to a
               random permutation.
        :param only_inference: Indicates if the model should be loaded to only restore inference capabilities or also
               training capabilities. Note that the training capabilities are currently not being fully restored.
        :param seed: Seed that is used for the prediction. Allows for a deterministic behavior of the predictions.
        :param batch_size_inference: This parameter is a trade-off between performance and memory consumption.
               The computation done with different values for batch_size_inference is the same,
               but it is split into smaller/larger batches.
        :param no_grad: If set to false, allows for the computation of gradients with respect to X_train and X_test.
               For this to correctly function no_preprocessing_mode must be set to true.
        """

        # Model file specification (Model name, Epoch)
        i = 0
        model_key = model_string + "|" + str(device)
        if model_string in self.models_in_memory:
            model, c, results_file = self.models_in_memory[model_key]
        else:
            model, c, results_file = load_model_workflow(
                i,
                -1,
                add_name=model_string,
                base_path=base_path,
                device=device,
                eval_addition="",
                only_inference=only_inference,
            )
            self.models_in_memory[model_key] = (model, c, results_file)
            if len(self.models_in_memory) == 2:
                print(
                    "Multiple models in memory. This might lead to memory issues. Consider calling remove_models_from_memory()"
                )
        # style, temperature = self.load_result_minimal(style_file, i, e)
        self.device = device
        self.model = model
        self.c = c
        self.style = None
        self.temperature = None
        self.N_ensemble_configurations = N_ensemble_configurations
        self.base__path = base_path
        self.base_path = base_path
        self.i = i
        self.model_string = model_string

        self.max_num_features = self.c["num_features"]
        self.max_num_classes = self.c["max_num_classes"]
        self.differentiable_hps_as_style = self.c["differentiable_hps_as_style"]

        self.no_preprocess_mode = no_preprocess_mode
        self.feature_shift_decoder = feature_shift_decoder
        self.multiclass_decoder = multiclass_decoder
        self.only_inference = only_inference
        self.seed = seed
        self.no_grad = no_grad

        assert (
            self.no_preprocess_mode if not self.no_grad else True
        ), "If no_grad is false, no_preprocess_mode must be true, because otherwise no gradient can be computed."

        self.batch_size_inference = batch_size_inference

    def remove_models_from_memory(self):
        self.models_in_memory = {}

    def load_result_minimal(self, path, i, e):
        with open(path, "rb") as output:
            _, _, _, style, temperature, optimization_route = CustomUnpickler(output).load()

            return style, temperature

    def _validate_targets(self, y):
        y_ = column_or_1d(y, warn=True)
        check_classification_targets(y)
        cls, y = np.unique(y_, return_inverse=True)
        if len(cls) < 2:
            raise ValueError("The number of classes has to be greater than one; got %d class" % len(cls))

        self.classes_ = cls

        return np.asarray(y, dtype=np.float64, order="C")

    def fit(self, X, y, overwrite_warning=False):
        """
        Validates the training set and stores it.

        If clf.no_grad (default is True):
        X, y should be of type np.array
        else:
        X should be of type torch.Tensors (y can be np.array or torch.Tensor)
        """
        if self.no_grad:
            # Check that X and y have correct shape
            X, y = check_X_y(X, y, force_all_finite=False)
        # Store the classes seen during fit
        y = self._validate_targets(y)
        self.label_encoder = LabelEncoder()
        y = self.label_encoder.fit_transform(y)

        self.X_ = X
        self.y_ = y

        if X.shape[1] > self.max_num_features:
            raise ValueError("The number of features for this classifier is restricted to ", self.max_num_features)
        if len(np.unique(y)) > self.max_num_classes:
            raise ValueError("The number of classes for this classifier is restricted to ", self.max_num_classes)
        if X.shape[0] > 1024 and not overwrite_warning:
            raise ValueError(
                "⚠️ WARNING: TabPFN is not made for datasets with a trainingsize > 1024. Prediction might take a while, be less reliable. We advise not to run datasets > 10k samples, which might lead to your machine crashing (due to quadratic memory scaling of TabPFN). Please confirm you want to run by passing overwrite_warning=True to the fit function."
            )

        # Return the classifier
        return self

    def predict_embeds(self, X, normalize_with_test=False, return_logits=False):
        """
        Predict the probabilities for the input X depending on the training set previously passed in the method fit.

        If no_grad is true in the classifier the function takes X as a numpy.ndarray. If no_grad is false X must be a
        torch tensor and is not fully checked.
        """
        # Check is fit had been called
        check_is_fitted(self)

        # Input validation
        if self.no_grad:
            # print(type(X))
            # if torch.is_tensor(X):
            #    X=X.detach().cpu().numpy()
            #    print(X.shape)
            # else:
            #    print(X.size())
            X = check_array(X, force_all_finite=False)
            X_full = np.concatenate([self.X_, X], axis=0)
            X_full = torch.tensor(X_full, device=self.device).float().unsqueeze(1)
        else:
            assert torch.is_tensor(self.X_) & torch.is_tensor(X), (
                "If no_grad is false, this function expects X as " "a tensor to calculate a gradient"
            )
            X_full = torch.cat((self.X_, X), dim=0).float().unsqueeze(1).to(self.device)

            if int(torch.isnan(X_full).sum()):
                print("X contains nans and the gradient implementation is not designed to handel nans.")

        y_full = np.concatenate([self.y_, np.zeros(shape=X.shape[0])], axis=0)
        y_full = torch.tensor(y_full, device=self.device).float().unsqueeze(1)

        eval_pos = self.X_.shape[0]

        prediction = transformer_predict(
            self.model[2],
            X_full,
            y_full,
            eval_pos,
            device=self.device,
            style=self.style,
            inference_mode=True,
            preprocess_transform="none" if self.no_preprocess_mode else "mix",
            normalize_with_test=normalize_with_test,
            N_ensemble_configurations=self.N_ensemble_configurations,
            softmax_temperature=self.temperature,
            multiclass_decoder=self.multiclass_decoder,
            feature_shift_decoder=self.feature_shift_decoder,
            differentiable_hps_as_style=self.differentiable_hps_as_style,
            seed=self.seed,
            return_logits=return_logits,
            no_grad=self.no_grad,
            batch_size_inference=self.batch_size_inference,
            **get_params_from_config(self.c),
        )
        prediction_, y_ = prediction.squeeze(0), y_full.squeeze(1).long()[eval_pos:]

        return prediction_

    def predict(self, X, return_winning_probability=False, normalize_with_test=False):
        p = self.predict_embeds(X, normalize_with_test=normalize_with_test)
        y = np.argmax(p, axis=-1)
        y = self.classes_.take(np.asarray(y, dtype=np.intp))
        if return_winning_probability:
            return y, p.max(axis=-1)
        return y


import time


def transformer_predict(
    model,
    eval_xs,
    eval_ys,
    eval_position,
    device="cpu",
    max_features=100,
    style=None,
    inference_mode=False,
    num_classes=2,
    extend_features=True,
    normalize_with_test=False,
    normalize_to_ranking=False,
    softmax_temperature=0.0,
    multiclass_decoder="permutation",
    preprocess_transform="mix",
    categorical_feats=[],
    feature_shift_decoder=False,
    N_ensemble_configurations=10,
    batch_size_inference=16,
    differentiable_hps_as_style=False,
    average_logits=True,
    fp16_inference=False,
    normalize_with_sqrt=False,
    seed=0,
    no_grad=True,
    return_logits=False,
    **kwargs,
):
    """

    :param model:
    :param eval_xs:
    :param eval_ys:
    :param eval_position:
    :param rescale_features:
    :param device:
    :param max_features:
    :param style:
    :param inference_mode:
    :param num_classes:
    :param extend_features:
    :param normalize_to_ranking:
    :param softmax_temperature:
    :param multiclass_decoder:
    :param preprocess_transform:
    :param categorical_feats:
    :param feature_shift_decoder:
    :param N_ensemble_configurations:
    :param average_logits:
    :param normalize_with_sqrt:
    :param metric_used:
    :return:
    """
    num_classes = len(torch.unique(eval_ys))

    def predict(eval_xs, eval_ys, used_style, softmax_temperature, return_logits):
        # Initialize results array size S, B, Classes

        # no_grad disables inference_mode, because otherwise the gradients are lost
        inference_mode_call = torch.inference_mode() if inference_mode and no_grad else NOP()
        with inference_mode_call:
            start = time.time()

            output = model.forward(
                (used_style.repeat(eval_xs.shape[1], 1) if used_style is not None else None, eval_xs, eval_ys.float()),
                single_eval_pos=eval_position
            )  # [:, :, 0:num_classes]
            # print("OUTPU",output.detach().cpu().numpy().shape)

            # output = output[:, :, 0:num_classes] / torch.exp(softmax_temperature)
            # if not return_logits:
            #    output = torch.nn.functional.softmax(output, dim=-1)
            # else:
            #    output[:, :, 1] = model((style.repeat(eval_xs.shape[1], 1) if style is not None else None, eval_xs, eval_ys.float()),
            #               single_eval_pos=eval_position)

            #    output[:, :, 1] = torch.sigmoid(output[:, :, 1]).squeeze(-1)
            #    output[:, :, 0] = 1 - output[:, :, 1]

        # print('RESULTS', eval_ys.shape, torch.unique(eval_ys, return_counts=True), output.mean(axis=0))

        return output

    def preprocess_input(eval_xs, preprocess_transform):
        import warnings

        if eval_xs.shape[1] > 1:
            raise Exception("Transforms only allow one batch dim - TODO")
        if preprocess_transform != "none":
            if preprocess_transform == "power" or preprocess_transform == "power_all":
                pt = PowerTransformer(standardize=True)
            elif preprocess_transform == "quantile" or preprocess_transform == "quantile_all":
                pt = QuantileTransformer(output_distribution="normal")
            elif preprocess_transform == "robust" or preprocess_transform == "robust_all":
                pt = RobustScaler(unit_variance=True)

        # eval_xs, eval_ys = normalize_data(eval_xs), normalize_data(eval_ys)
        eval_xs = normalize_data(eval_xs, normalize_positions=-1 if normalize_with_test else eval_position)

        # Removing empty features
        eval_xs = eval_xs[:, 0, :]
        sel = [len(torch.unique(eval_xs[0 : eval_ys.shape[0], col])) > 1 for col in range(eval_xs.shape[1])]
        eval_xs = eval_xs[:, sel]

        warnings.simplefilter("error")
        if preprocess_transform != "none":
            eval_xs = eval_xs.cpu().numpy()
            feats = (
                set(range(eval_xs.shape[1]))
                if "all" in preprocess_transform
                else set(range(eval_xs.shape[1])) - set(categorical_feats)
            )
            for col in feats:
                try:
                    pt.fit(eval_xs[0:eval_position, col : col + 1])
                    trans = pt.transform(eval_xs[:, col : col + 1])
                    # print(scipy.stats.spearmanr(trans[~np.isnan(eval_xs[:, col:col+1])], eval_xs[:, col:col+1][~np.isnan(eval_xs[:, col:col+1])]))
                    eval_xs[:, col : col + 1] = trans
                except:
                    pass
            eval_xs = torch.tensor(eval_xs).float()
        warnings.simplefilter("default")

        eval_xs = eval_xs.unsqueeze(1)

        # TODO: Caution there is information leakage when to_ranking is used, we should not use it
        # print("NORMALIZE? ", normalize_with_test)
        eval_xs = (
            remove_outliers(eval_xs, normalize_positions=-1 if normalize_with_test else eval_position)
            if not normalize_to_ranking
            else normalize_data(to_ranking_low_mem(eval_xs))
        )
        # Rescale X
        eval_xs = normalize_by_used_features_f(
            eval_xs, eval_xs.shape[-1], max_features, normalize_with_sqrt=normalize_with_sqrt
        )

        return eval_xs.to(device)

    eval_xs, eval_ys = eval_xs.to(device), eval_ys.to(device)
    eval_ys = eval_ys[:eval_position]

    model.to(device)

    model.eval()

    import itertools

    if not differentiable_hps_as_style:
        style = None

    if style is not None:
        style = style.to(device)
        style = style.unsqueeze(0) if len(style.shape) == 1 else style
        num_styles = style.shape[0]
        softmax_temperature = (
            softmax_temperature if softmax_temperature.shape else softmax_temperature.unsqueeze(0).repeat(num_styles)
        )
    else:
        num_styles = 1
        style = None
        softmax_temperature = torch.log(torch.tensor([0.8]))

    styles_configurations = range(0, num_styles)

    def get_preprocess(i):
        if i == 0:
            return "power_all"
        #            if i == 1:
        #                return 'robust_all'
        if i == 1:
            return "none"

    preprocess_transform_configurations = (
        ["none", "power_all"] if preprocess_transform == "mix" else [preprocess_transform]
    )

    if seed is not None:
        torch.manual_seed(seed)

    feature_shift_configurations = torch.randperm(eval_xs.shape[2]) if feature_shift_decoder else [0]
    class_shift_configurations = (
        torch.randperm(len(torch.unique(eval_ys))) if multiclass_decoder == "permutation" else [0]
    )

    ensemble_configurations = list(itertools.product(class_shift_configurations, feature_shift_configurations))
    # default_ensemble_config = ensemble_configurations[0]

    rng = random.Random(seed)
    rng.shuffle(ensemble_configurations)
    ensemble_configurations = list(
        itertools.product(ensemble_configurations, preprocess_transform_configurations, styles_configurations)
    )
    ensemble_configurations = ensemble_configurations[0:N_ensemble_configurations]
    # if N_ensemble_configurations == 1:
    #    ensemble_configurations = [default_ensemble_config]

    output = None

    eval_xs_transformed = {}
    inputs, labels = [], []
    start = time.time()
    for ensemble_configuration in ensemble_configurations:
        (
            (class_shift_configuration, feature_shift_configuration),
            preprocess_transform_configuration,
            styles_configuration,
        ) = ensemble_configuration

        style_ = style[styles_configuration : styles_configuration + 1, :] if style is not None else style
        softmax_temperature_ = softmax_temperature[styles_configuration]

        eval_xs_, eval_ys_ = eval_xs.clone(), eval_ys.clone()

        if preprocess_transform_configuration in eval_xs_transformed:
            eval_xs_ = eval_xs_transformed[preprocess_transform_configuration].clone()
        else:
            eval_xs_ = preprocess_input(eval_xs_, preprocess_transform=preprocess_transform_configuration)
            if no_grad:
                eval_xs_ = eval_xs_.detach()
            eval_xs_transformed[preprocess_transform_configuration] = eval_xs_

        eval_ys_ = ((eval_ys_ + class_shift_configuration) % num_classes).float()

        eval_xs_ = torch.cat(
            [eval_xs_[..., feature_shift_configuration:], eval_xs_[..., :feature_shift_configuration]], dim=-1
        )

        # Extend X
        if extend_features:
            eval_xs_ = torch.cat(
                [
                    eval_xs_,
                    torch.zeros((eval_xs_.shape[0], eval_xs_.shape[1], max_features - eval_xs_.shape[2])).to(device),
                ],
                -1,
            )
        inputs += [eval_xs_]
        labels += [eval_ys_]

    inputs = torch.cat(inputs, 1)
    inputs = torch.split(inputs, batch_size_inference, dim=1)
    labels = torch.cat(labels, 1)
    labels = torch.split(labels, batch_size_inference, dim=1)
    # print('PREPROCESSING TIME', str(time.time() - start))
    outputs = []
    start = time.time()
    for batch_input, batch_label in zip(inputs, labels):
        # preprocess_transform_ = preprocess_transform if styles_configuration % 2 == 0 else 'none'
        import warnings

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="None of the inputs have requires_grad=True. Gradients will be None"
            )
            warnings.filterwarnings(
                "ignore",
                message="torch.cuda.amp.autocast only affects CUDA ops, but CUDA is not available.  Disabling.",
            )
            if device == "cpu":
                output_batch = checkpoint(predict, batch_input, batch_label, style_, softmax_temperature_, True)
            else:
                with torch.cuda.amp.autocast(enabled=fp16_inference):
                    output_batch = checkpoint(predict, batch_input, batch_label, style_, softmax_temperature_, True)
        outputs += [output_batch]
    # print('MODEL INFERENCE TIME ('+str(batch_input.device)+' vs '+device+', '+str(fp16_inference)+')', str(time.time()-start))
    # print("OUTPU",outputs)
    outputs = torch.cat(outputs, 1)
    return outputs


def get_params_from_config(c):
    return {
        "max_features": c["num_features"],
        "rescale_features": c["normalize_by_used_features"],
        "normalize_to_ranking": c["normalize_to_ranking"],
        "normalize_with_sqrt": c.get("normalize_with_sqrt", False),
    }



import pytest
from sklearn.ensemble import RandomForestClassifier

from alpbench.benchmark.BenchmarkConnector import DataFileBenchmarkConnector
from alpbench.evaluation.experimenter.DefaultSetup import ensure_default_setup
from alpbench.pipeline.QueryStrategy import RandomQueryStrategy

base_folder = "temp/"
learner_file = base_folder + "learner.json"
query_strategy_file = base_folder + "query_strategy.json"
scenario_file = base_folder + "scenario.json"
setting_file = base_folder + "setting.json"


@pytest.fixture
@pytest.mark.usefixtures("scenario")
def benchmark_connector(scenario):
    bc = DataFileBenchmarkConnector(
        learner_file=learner_file,
        query_strategy_file=query_strategy_file,
        scenario_file=scenario_file,
        setting_file=setting_file,
    )
    ensure_default_setup(bc)
    bc.load_or_create_scenario(
        scenario.openml_id,
        test_split_seed=scenario.test_split_seed,
        train_split_seed=scenario.train_split_seed,
        seed=scenario.seed,
        setting_id=scenario.setting.setting_id,
    )
    bc.dump()
    return DataFileBenchmarkConnector(
        learner_file=learner_file,
        query_strategy_file=query_strategy_file,
        scenario_file=scenario_file,
        setting_file=setting_file,
    )


def test_load_learner(benchmark_connector):
    learner = benchmark_connector.load_learner(1)
    assert learner is not None


def test_load_learner_by_name(benchmark_connector):
    learner = benchmark_connector.load_learner_by_name("svm_lin")
    assert learner is not None


def test_load_query_strategy(benchmark_connector):
    query_strategy = benchmark_connector.load_query_strategy(1)
    assert query_strategy is not None


def test_load_query_strategy_by_name(benchmark_connector):
    query_strategy = benchmark_connector.load_query_strategy_by_name("random")
    assert query_strategy is not None


def test_load_setting(benchmark_connector):
    setting = benchmark_connector.load_setting(1)
    assert setting is not None


def test_load_setting_by_name(benchmark_connector):
    setting = benchmark_connector.load_setting_by_name("small")
    assert setting is not None


def test_load_scenario(benchmark_connector):
    scenario = benchmark_connector.load_scenario(1)
    assert scenario is not None


def test_fail_load_nonexistent_setting(benchmark_connector):
    with pytest.raises(BaseException):
        benchmark_connector.load_setting(-1)


def test_fail_load_nonexistent_scenario(benchmark_connector):
    with pytest.raises(BaseException):
        benchmark_connector.load_scenario(-1)


def test_fail_load_nonexistent_setting_by_name(benchmark_connector):
    with pytest.raises(BaseException):
        benchmark_connector.load_setting_by_name("foe")


def test_create_new_setting(benchmark_connector):
    setting = benchmark_connector.load_or_create_setting("test", 0.5, "absolute", 0.3, 2, 2, 3)
    assert setting.get_setting_id() == 6


def test_fail_load_nonexistent_learner(benchmark_connector):
    with pytest.raises(BaseException):
        benchmark_connector.load_learner_by_name("foe")


def test_fail_load_nonexistent_learner_by_id(benchmark_connector):
    with pytest.raises(BaseException):
        benchmark_connector.load_learner(-1)


def test_load_new_learner(benchmark_connector):
    benchmark_connector.load_or_create_learner("foe", RandomForestClassifier(n_estimators=42))


def test_fail_load_nonexistent_query_strategy_by_name(benchmark_connector):
    with pytest.raises(BaseException):
        benchmark_connector.load_query_strategy_by_name("foe")


def test_fail_load_nonexistent_query_strategy(benchmark_connector):
    with pytest.raises(BaseException):
        benchmark_connector.load_query_strategy(-1)


def test_load_new_query_strategy(benchmark_connector):
    benchmark_connector.load_or_create_query_strategy("foe", RandomQueryStrategy(seed=10))


def test_init_default_file_paths():
    dfbc = DataFileBenchmarkConnector()
    assert dfbc.learner_file != DataFileBenchmarkConnector.base_folder + DataFileBenchmarkConnector.learner_file
    dfbc.cleanup()



from sklearn.ensemble import RandomForestClassifier

from alpbench.util.common import format_insert_query, format_select_query, fullname, instantiate_class_by_fqn


def test_insert_query():
    query = format_insert_query("TBL", {"x": 1})
    assert query == "INSERT INTO TBL (x) VALUES ('1');"


def test_select_query():
    query = format_select_query("TBL", where={"x": 1})
    assert query == "SELECT * FROM TBL WHERE x='1';"

    query = format_select_query("TBL")
    assert query == "SELECT * FROM TBL;"


def test_instantiate_class_by_fqn():
    rf = instantiate_class_by_fqn("sklearn.ensemble.RandomForestClassifier")
    assert type(rf) is RandomForestClassifier


def test_fullname_of_builtin():
    assert fullname("a") == "builtins.str"



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.colors import Normalize, LinearSegmentedColormap

import pandas as pd
from py_experimenter.experimenter import PyExperimenter
import json
import os
from scipy.stats import ttest_ind_from_stats
from scipy.stats import ttest_ind as tt



class BudgetPerformancePlot:
    """BudgetPerformancePlot

    This class plots the performance for a given learning algorithm and openmlid of different query strategies over
    the whole active learning procedure resulting in so-calles area under the budget curves (AUBC).

    Args:
        df (pd.DataFrame): The dataframe.
        openml_id (int): The openml id.
        learner_name (str): The learner name.
        metric (str): The metric.
        path_to_save (str): The path to save the plot.

    Attributes:
        df (pd.DataFrame): The dataframe.
        openml_id (int): The openml id.
        learner_name (str): The learner name.
        metric (str): The metric.
        path_to_save (str): The path to save the plot.
        plot_data (dict): The data to plot.
        num_seeds (int): The number of seeds (used to determine the std error).


    """
    def __init__(self, df, openml_id, learner_name, metric, path_to_save=None):
        self.df = df
        self.openml_id = openml_id
        self.learner_name = learner_name
        self.metric = metric
        self.path_to_save = path_to_save
        self.plot_data = None
        self.num_seeds = None

    def generate_plot_data(self):
        """
        This function generates the data to plot.
        """
        # get data for openml_id
        df = self.df[self.df['openml_id'] == self.openml_id]
        # get data for learner
        df = df[df['learner_name'] == self.learner_name]
        # get unique query strategy names
        query_strategies = df['query_strategy_name'].unique()
        # get unique budget values
        budgets = df['len_X_l'].unique()
        # create dict to store data
        data = {}
        for qs in query_strategies:
            data[qs] = {'budget': [], 'mean': [], 'std': []}
            for enum,budget in enumerate(budgets):
                # get data for query strategy and budget
                df_temp = df[(df['query_strategy_name'] == qs) & (df['len_X_l'] == budget)]

                # reset index
                df_temp.reset_index()
                # get mean and std of metric
                mean = df_temp[self.metric].mean()
                std = df_temp[self.metric].std()

                df_reset = df_temp[self.metric].reset_index()
                # append data to dict
                data[qs]['budget'].append(budget)
                data[qs]['mean'].append(mean)
                data[qs]['std'].append(std)

        # get num of seeds
        self.num_seeds = len(df['seed'].unique())
        self.plot_data = data


    def show(self, show_fig = False):
        """
        This function plots the performance of different query strategies over the budget and saves
        it as .pdf under the specified path.
        """
        data = self.plot_data
        if len(data.keys()) == 0:
            return
        else:
            oid = self.openml_id
            # we want to order the QS by grouping them into categories, also each category gets a different
            # color coding (uncertainty based are redish, representative are greenish, hybrid are blueish,
            # random is pink)
            keys = data.keys()
            list_of_qs = list(keys)
            all_qs_ordered = ["random", "entropy", "least_confident", "margin", "power_margin",
                        "max_entropy", "bald", "power_bald", "qbc_variance_ratio",
                        "kmeans", "core_set","typ_cluster",
                        "cluster_margin", "falcun",  "weighted_cluster"]
            # order the query strategies
            ordered_qs = []
            for qs in all_qs_ordered:
                if qs in list_of_qs:
                    ordered_qs.append(qs)

            fig, ax = plt.subplots(1)
            # colormap of len(keys)
            colors = plt.cm.tab20(np.linspace(0, 1, len(keys)))
            color_dict = {"random": "magenta", "least_confident": "rosybrown", "margin": "red",
                          "entropy": "orange", "power_margin": "brown", "bald": "tomato",
                          "power_bald": "coral", "max_entropy": "sandybrown", "qbc_variance_ratio": "peachpuff",
                          "core_set": "limegreen", "typ_cluster": "forestgreen", "cluster_margin": "mediumblue",
                          "weighted_cluster": "turquoise", "falcun": "blue"}
            for cl, key in enumerate(ordered_qs):
                budget = np.array(data[key]['budget'])
                mu = np.array(data[key]['mean'])
                # std error
                std = np.array(data[key]['std'])/np.sqrt(self.num_seeds)
                qs = key
                cl = color_dict[qs]
                plt.plot(budget, mu, lw=2, label=key, color=cl)
                plt.fill_between(budget, mu+std, mu-std,  facecolor=cl, alpha=0.5)


            qs_for_figure = {"margin": "MS", "least_confident": "LC", "entropy": "ES",
                                "power_margin": "PowMS", "bald": "BALD", "power_bald": "PowBALD",
                                "max_entropy": "MaxEnt", "qbc_variance_ratio": "QBC VR",
                                "core_set": "CoreSet", "typ_cluster": "TypClu", "cluster_margin": "CluMS",
                                "weighted_cluster": "Clue", "falcun": "FALCUN", "random": "Rand"}
            learner_for_figure = {"knn_3": "KNN", "svm_rbf": "SVM (rbf)", "mlp": "MLP", "rf_entropy": "RF (entr)",
                                  "rf_gini": "RF (gini)", "svm_lin": "SVM (lin)", "catboost": "Catboost", "xgb": "XGB",
                                  "tabnet": "Tabnet", "tabpfn": "TabPFN"}

            handles, labels = plt.gca().get_legend_handles_labels()

            new_labels = []
            for label in labels:
                new_labels.append(qs_for_figure[label])

            plt.title(learner_for_figure[self.learner_name] + " on id " + str(self.openml_id), fontsize=30)
            plt.legend(handles, new_labels, fontsize=20, loc='lower right')# fontsize=25,loc='center left', bbox_to_anchor=(1, 0.5))
            plt.yticks(fontsize=20)
            plt.xticks(fontsize=20)
            ax.set_xlabel('Number of labeled instances', fontsize=25)
            ax.set_ylabel('test accuracy', fontsize=25)
            [l.set_visible(False) for (i, l) in enumerate(ax.xaxis.get_ticklabels()) if i % 2 != 0]

            # save image
            if self.path_to_save is not None:
                if not os.path.exists(self.path_to_save):
                    os.makedirs(self.path_to_save)
            else:
                self.path_to_save = "FIGURES/BUDGET_PERFORMANCE_PLOT/" +  str(self.openml_id) + "/"
                if not os.path.exists(self.path_to_save):
                    os.makedirs(self.path_to_save)
            self.path_to_save = self.path_to_save + str(self.learner_name) + ".pdf"
            fig.savefig(self.path_to_save, facecolor='white', transparent=True, bbox_inches='tight')
            if show_fig:
                plt.show()
            plt.close()


from alpbench.benchmark.Observer import StatisticalPerformanceObserver


class LogTableObserver(StatisticalPerformanceObserver):
    """LogTableObserver

    This class observes the performance of the model and the labeling process and logs the results in the database.

    Args:
        StatisticalPerformanceObserver (class): The parent class of the LogTableObserver.
        result_processor (class): The result processor class.
        X_test (array): The test data.
        y_test (array): The test labels.

    Attributes:
        model_performance_tbl (str): The name of the table where the model performance is logged.
        labeling_log_tbl (str): The name of the table where the labeling process is logged.
    """

    model_performance_tbl = "accuracy_log"
    labeling_log_tbl = "labeling_log"

    def __init__(self, result_processor, X_test, y_test):
        super().__init__(X_test, y_test)
        self.result_processor = result_processor

    def observe_data(self, iteration, X_u_selected, y_u_selected, X_l_aug, y_l_aug, X_u_red):
        """
        Computes labeling statistics and log the results in the database.

        Parameters:
            iteration (int): The current iteration.
            X_u_selected (array): The selected unlabeled data.
            y_u_selected (array): The selected unlabeled labels.
            X_l_aug (array): The augmented labeled data.
            y_l_aug (array): The augmented labeled labels.
            X_u_red (array): The reduced unlabeled data.

        Returns:
            None
        """
        eval_scores = super().compute_labeling_statistics(
            iteration, X_u_selected, y_u_selected, X_l_aug, y_l_aug, X_u_red
        )
        self.result_processor.process_logs({LogTableObserver.labeling_log_tbl: eval_scores})

    def observe_model(self, iteration, model):
        """
        Computes model performances and log the results in the database.

        Parameters:
            iteration (int): The current iteration.
            model (class): The model

        Returns:
            None
        """
        eval_scores = super().compute_model_performances(iteration, model)
        self.result_processor.process_logs({LogTableObserver.model_performance_tbl: eval_scores})


class SparseLogTableObserver(StatisticalPerformanceObserver):
    """SparseLogTableObserver

    This class observes the performance of the model and the labeling process and logs the results in the database.
    To reduce the number of logs, the results are only logged after a whole active learning procedure is finished.

    Args:
        StatisticalPerformanceObserver (class): The parent class of the LogTableObserver.
        result_processor (class): The result processor class.
        X_test (array): The test data.
        y_test (array): The test labels.

    Attributes:
        model_performance_tbl (str): The name of the table where the model performance is logged.
        labeling_log_tbl (str): The name of the table where the labeling process is logged.
    """
    model_performance_tbl = "accuracy_log"
    labeling_log_tbl = "labeling_log"

    def __init__(self, result_processor, X_test, y_test):
        super().__init__(X_test, y_test)
        self.result_processor = result_processor

    def observe_data(self, iteration, X_u_selected, y_u_selected, X_l_aug, y_l_aug, X_u_red):
        """
        Computes labeling statistics per iteration.

        Parameters:
            iteration (int): The current iteration.
            X_u_selected (array): The selected unlabeled data.
            y_u_selected (array): The selected unlabeled labels.
            X_l_aug (array): The augmented labeled data.
            y_l_aug (array): The augmented labels.
            X_u_red (array): The reduced unlabeled data.

        Returns:
            eval_scores (dict): The evaluation scores.
        """
        eval_scores = super().compute_labeling_statistics(
            iteration, X_u_selected, y_u_selected, X_l_aug, y_l_aug, X_u_red
        )
        return eval_scores

    def observe_model(self, iteration, model):
        """
        Computes model performances per iteration.

        Parameters:
            iteration (int): The current iteration.
            model (class): The model

        Returns:
            eval_scores (dict): The evaluation scores.
        """
        eval_scores = super().compute_model_performances(iteration, model)
        return eval_scores

    def log_data(self, dict):
        """
        Logs the labeling statistics in the database after the active learning procedure is finished.

        Parameters:
            dict (dict): The evaluation scores.

        Returns:
            None
        """
        self.result_processor.process_logs({LogTableObserver.labeling_log_tbl: dict})

    def log_model(self, dict):
        """
        Logs the model performances in the database after the active learning procedure is finished.

        Parameters:
            dict (dict): The evaluation scores.

        Returns:
            None
        """
        self.result_processor.process_logs({LogTableObserver.model_performance_tbl: dict})






from alpbench.benchmark.BenchmarkSuite import OpenMLBenchmarkSuite, TabZillaBenchmarkSuite


def test_openml_benchmark_suite_loader():
    dids = OpenMLBenchmarkSuite(openml_benchmark_id=99).get_openml_dataset_ids()
    assert len(dids) == 72


def test_tabzilla_benchmark_suite_loader():
    dids = TabZillaBenchmarkSuite().get_openml_dataset_ids()
    assert len(dids) == 34



import json
import os
import numpy as np
import pandas as pd
import pandas as pd
from py_experimenter.experimenter import PyExperimenter
from scipy.stats import ttest_ind_from_stats
from scipy.stats import ttest_ind as tt
from get_results import StudyData



# DEFINE METRIC, DATABASENAME, GENERATE PANDAS DATAFRAME AND SAVE IT
grid = {}
metric = ["test_accuracy"]

db_name = "alpbench_ecai"
setting_string = "small"
db_config_file = "../../evaluation/experimenter/config/db_conf.yml"
exp_scenario_file = "../../evaluation/experimenter/config/experiments.yml"
grid["db_name"] = db_name
grid["setting_name"] = setting_string


source = db_config_file, exp_scenario_file
studyData = StudyData(source=source, grid=grid, csv_path=f"results/dataframes/")
end_df = studyData.get_aubc_df()
# check if path exists
directory_path = "DATAFRAMES/" + setting_string + "/"
if not os.path.exists(directory_path):
    print("does not exist")
    os.makedirs(directory_path)
end_df.to_csv(directory_path + db_name +"_"+ setting_string + "_dynamic.csv")



#learners = ["rf_entropy", "tabpfn", "catboost"]
#query_strategies = ["margin", "entropy", "cluster_margin", "typ_cluster", "random"]


class ActiveLearningSetting:
    """Active Learning Setting

    The active learning setting defines constraints and design choices of one active learning setup. This involves the
    size of the labeled training data, the size of the test data, the number of iterations, the number of samples
    queried per iteration, and a task-dependent factor in case a dynamic setting is considered (i.e. number of samples
    queried depends on the number of classes of the given dataset).

    Args:
        setting_id (int): id of the setting in the database
        setting_name (str): descriptor of the setting
        setting_labeled_train_size (float): size of the labeled training size
        setting_train_type (str): type of the size parameter: number of data points or share of the (training) dataset
        setting_test_size (float): size of the test data (always given as a share of the full dataset)
        number_of_iterations (int): number of iterations
        number_of_queries (int): number of queries queried per iteration
        factor (int): task-dependent factor

    Attributes:
        setting_id (int): id of the setting in the database
        setting_name (str): descriptor of the setting
        setting_labeled_train_size (float): size of the labeled training size
        setting_train_type (str): type of the size parameter: number of data points or share of the (training) dataset
        setting_test_size (float): size of the test data (always given as a share of the full dataset)
        number_of_iterations (int): number of iterations
        number_of_queries (int): number of queries queried per iteration
        factor (int): task-dependent factor
    """

    def __init__(
        self,
        setting_id,
        setting_name,
        setting_labeled_train_size,
        setting_train_type,
        setting_test_size,
        number_of_iterations,
        number_of_queries,
        factor,
    ):
        # id of the setting in the database
        self.setting_id = setting_id
        # descriptor of the setting
        self.setting_name = setting_name
        # size of the labeled training size
        self.setting_labeled_train_size = float(setting_labeled_train_size)
        # type of the size parameter: number of data points or share of the (training) dataset
        self.setting_train_type = setting_train_type
        # size of the test data (always given as a share of the full dataset)
        self.setting_test_size = float(setting_test_size)
        # number of iterations
        self.number_of_iterations = number_of_iterations
        # number of samples queried per iteration
        self.number_of_queries = number_of_queries
        # task-dependent factor
        self.factor = factor

    def get_setting_id(self):
        """
        Get the setting id.
        """
        return self.setting_id

    def get_setting_name(self):
        """
        Get the setting name.
        """
        return self.setting_name

    def get_factor(self):
        """
        Get the factor.
        """
        return self.factor

    def get_setting_labeled_train_size(self):
        """
        Get the size of the labeled training data.
        """
        return self.setting_labeled_train_size

    def get_setting_train_type(self):
        """
        Get the training type, absolute or relative.
        """
        return self.setting_train_type

    def get_setting_test_size(self):
        """
        Get the size of the test data.
        """
        return self.setting_test_size

    def get_number_of_iterations(self):
        """
        Get the number of iterations.
        """
        return self.number_of_iterations

    def get_number_of_queries(self):
        """
        Get the number of samples queried per iteration.
        """
        return self.number_of_queries

    def __repr__(self):
        return "<ActiveLearningSetting> " + str(self.__dict__)



import pytest

from .fixtures.scenario import (
    FACTOR,
    NUMBER_OF_IT,
    NUMBER_OF_QUERIES,
    SETTING_ID,
    SETTING_NAME,
    SETTING_TEST_SIZE,
    SETTING_TRAIN_SIZE,
    SETTING_TRAIN_TYPE,
)


@pytest.mark.usefixtures("setting")
def test_get_setting_id(setting):
    assert setting.get_setting_id() == SETTING_ID


@pytest.mark.usefixtures("setting")
def test_get_setting_name(setting):
    assert setting.get_setting_name() == SETTING_NAME


@pytest.mark.usefixtures("setting")
def test_get_setting_labeled_train_size(setting):
    assert setting.get_setting_labeled_train_size() == SETTING_TRAIN_SIZE


@pytest.mark.usefixtures("setting")
def test_get_setting_train_type(setting):
    assert setting.get_setting_train_type() == SETTING_TRAIN_TYPE


@pytest.mark.usefixtures("setting")
def test_get_setting_test_size(setting):
    assert setting.get_setting_test_size() == SETTING_TEST_SIZE


@pytest.mark.usefixtures("setting")
def test_get_number_of_iterations(setting):
    assert setting.get_number_of_iterations() == NUMBER_OF_IT


@pytest.mark.usefixtures("setting")
def test_get_number_of_samples(setting):
    assert setting.get_number_of_queries() == NUMBER_OF_QUERIES


@pytest.mark.usefixtures("setting")
def test_get_factor(setting):
    assert setting.get_factor() == FACTOR


@pytest.mark.usefixtures("setting")
def test_repr(setting):
    assert str(setting).startswith("<ActiveLearningSetting>")



import pytest

from alpbench.pipeline.ActiveLearningPipeline import ActiveLearningPipeline
from alpbench.pipeline.Oracle import Oracle
from alpbench.pipeline.QueryStrategy import RandomQueryStrategy
from alpbench.util.transformer_prediction_interface import TabPFNClassifier


@pytest.mark.usefixtures("scenario")
def test_tabpfn_executions(scenario):
    print("get data split")
    X_l, y_l, X_u, y_u, X_test, y_test = scenario.get_data_split()

    print("setup learner")
    learner = TabPFNClassifier(device="cpu", N_ensemble_configurations=32)
    query_strategy = RandomQueryStrategy(42)

    ALP = ActiveLearningPipeline(
        learner=learner,
        query_strategy=query_strategy,
        init_budget=10,
        num_iterations=1,
        num_queries_per_iteration=10,
    )

    oracle = Oracle(X_u, y_u)
    ALP.active_fit(X_l, y_l, X_u, oracle)
    y_test_hat = ALP.predict(X_test)
    assert len(y_test_hat) == len(y_test)
